import logging
from dataclasses import dataclass

import numpy as np
import pandas as pd
import xarray as xr

from power_perceiver.load_prepared_batches.data_sources import (
    PV,
    HRVSatellite,
    PreparedDataSource,
    XarrayBatch,
)
from power_perceiver.load_prepared_batches.data_sources.pv import apply_pv_mask

_log = logging.getLogger(__name__)


@dataclass
class SelectPVSystemsNearCenterOfImage:
    """Selects PV systems near the top-middle of the imagery.

    The returned XarrayBatch will have the same number of spaces for PV systems as the input
    XarrayBatch, but the PV systems which are unselected will be set to false in the pv_mask.

    Initialisation args:
        image_data_loader_class: The name of the data source which defines the geospatial
            boundaries we'll use to select PV systems.
        geo_border_m: When selecting a single PV system from the top-middle of the HRV satellite
            imagery, use geo_border_m to define how much to reduce the selection rectangle
            before selecting PV systems. In meters.
        drop_examples: If True then drop examples (from all data sources) which have no PV systems.
    """

    image_data_loader_class: PreparedDataSource = HRVSatellite
    geo_border_m: pd.Series = pd.Series(dict(left=8_000, right=8_000, bottom=32_000, top=16_000))
    drop_examples: bool = True

    def __call__(self, xr_batch: XarrayBatch) -> XarrayBatch:
        image_dataset = xr_batch[self.image_data_loader_class]
        pv_dataset = xr_batch[PV]
        pv_id_indexes_for_all_examples = []
        # If this loop is too slow then it may be possible to vectorise this code.
        for example_i in image_dataset.example:
            # Get inner rectangle for image dataset:
            image_dataset_for_example = image_dataset.sel(example=example_i)
            inner_rectangle = self._get_maximal_regular_inner_rectangle(image_dataset_for_example)

            # Reduce inner rectangle:
            inner_rectangle["left"] += self.geo_border_m["left"]
            inner_rectangle["right"] -= self.geo_border_m["right"]
            inner_rectangle["bottom"] += self.geo_border_m["bottom"]
            inner_rectangle["top"] -= self.geo_border_m["top"]

            # Sanity check:
            if any(inner_rectangle.isnull()):
                msg = f"At least one inner_rectangle value is NaN!\n{inner_rectangle=}"
                _log.error(msg)
                raise ValueError(msg)

            # Find PV systems within the inner rectangle:
            pv_dataset_for_example = pv_dataset.sel(example=example_i)
            pv_locations = pv_dataset_for_example[["x_osgb", "y_osgb"]]
            pv_system_selection_mask = (
                (pv_locations.x_osgb >= inner_rectangle["left"])
                & (pv_locations.x_osgb <= inner_rectangle["right"])
                & (pv_locations.y_osgb >= inner_rectangle["bottom"])
                & (pv_locations.y_osgb <= inner_rectangle["top"])
            )
            pv_id_indexes_for_all_examples.append(pv_system_selection_mask)

        # Set PV systems outside of the inner_rectangle to NaN.
        mask = xr.concat(pv_id_indexes_for_all_examples, dim="example")
        xr_batch[PV]["pv_mask"] = pv_dataset.pv_mask & mask
        xr_batch[PV] = apply_pv_mask(xr_batch[PV])

        if self.drop_examples:
            # Drop examples which don't have any PV systems.
            n_pv_systems_per_example = mask.sum(dim="pv_system")
            examples_to_drop = np.where(n_pv_systems_per_example == 0)[0]
            for data_loader_class, xr_dataset in xr_batch.items():
                xr_batch[data_loader_class] = xr_dataset.drop_sel(example=examples_to_drop)

        return xr_batch

    def _get_maximal_regular_inner_rectangle(self, image_dataset: xr.Dataset) -> pd.Series:
        """Find the geographical boundaries (in OSGB coordinates) of the regular inner rectangle.

        Getting the OSGB boundaries for the PV systems is a bit fiddly because the satellite image
        is a skewed rectangle. We want the regular rectangle that fits within the OSGB boundaries:

            skewed satellite image                # noqa W605
            \--------\
             \         \-------\
              \    |===========|\
               \   | selection | \
                \  |           |  \
                 \ |===========|   \
                  \--------\        \
                            \--------\

        And the skew angle and direction might vary (e.g. if we compare images from Italy vs UK).
        So, to find the left value for _x_, we find the left-most coordinate for the top and
        bottom rows satellite image, and find the _max_ (right-most) value of this pair of numbers.

        Args:
            image_dataset: xarray Dataset for a single example. Must have these coordinates:
                y_osgb, x_osgb

        Returns:
            pd.Series with keys left, right, top, bottom. The values are the OSGB coordinates of
              the relevant axis. e.g. the left and right are the values for the x axis.
        """
        bounds = pd.Series(index=["left", "right", "top", "bottom"], dtype=np.float32)

        # Handle X coordinates. Get the x_osgb coords for the top and bottom rows of pixels.
        x_osgb = image_dataset["x_osgb"].isel(y=[0, -1])
        bounds["left"] = x_osgb.isel(x=0).max().values
        bounds["right"] = x_osgb.isel(x=-1).min().values

        # Handle Y coordinates. Get the y_osgb coords for the left and right columns of pixels.
        y_osgb = image_dataset["y_osgb"].isel(x=[0, -1])
        bounds["bottom"] = y_osgb.isel(y=0).max().values
        bounds["top"] = y_osgb.isel(y=-1).min().values

        # Sanity check!
        x_osgb_range = bounds["right"] - bounds["left"]
        y_osgb_range = bounds["top"] - bounds["bottom"]
        assert x_osgb_range > 0
        assert y_osgb_range > 0

        return bounds



from dataclasses import dataclass

import einops
import torch
from torch import nn

from power_perceiver.consts import BatchKey

N_PV_SYSTEMS_PER_EXAMPLE = 8


@dataclass(eq=False)
class PVRNN(nn.Module):
    hidden_size: int
    num_layers: int
    dropout: float

    def __post_init__(self):
        super().__init__()

        rnn_kwargs = dict(
            hidden_size=self.hidden_size,
            num_layers=self.num_layers,
            dropout=self.dropout,
            batch_first=True,
        )
        self.pv_rnn_history_encoder = nn.RNN(
            # Each timestep of the encoder RNN receives the output of the satellite_transformer
            # (which is of size d_model per PV system and per timestep), for a single PV system,
            # plus one timestep of that PV system's history.
            input_size=self.hidden_size + 1,
            **rnn_kwargs,
        )
        self.pv_rnn_future_decoder = nn.RNN(
            # Each timestep of the decoder RNN receives the output of the satellite_transformer
            # (which is of size d_model per PV system and per timestep) for a single PV system.
            input_size=self.hidden_size,
            **rnn_kwargs,
        )

    def forward(
        self, x: dict[BatchKey, torch.Tensor], sat_trans_pv_attn_out: torch.Tensor
    ) -> torch.Tensor:
        # Prepare inputs for the pv_rnn_history_encoder:
        # Each timestep of the encoder RNN receives the output of the satellite_transformer
        # (which is of size d_model per PV system and per timestep), for a single PV system,
        # plus one timestep of that PV system's history.
        # Some of `x[BatchKey.pv]` will be NaN. But that's OK. We mask missing examples.
        # And use nan_to_num later in this function.
        pv_t0_idx = x[BatchKey.pv_t0_idx]
        num_pv_timesteps = x[BatchKey.pv].shape[1]
        hist_pv = x[BatchKey.pv][:, : pv_t0_idx + 1]  # Shape: example, time, n_pv_systems

        # Reshape so each PV system is seen as a different example:
        hist_pv = einops.rearrange(
            hist_pv, "example time n_pv_systems -> (example n_pv_systems) time 1"
        )
        sat_trans_pv_attn_out = einops.rearrange(
            sat_trans_pv_attn_out,
            "example time n_pv_systems d_model -> (example n_pv_systems) time d_model",
            n_pv_systems=N_PV_SYSTEMS_PER_EXAMPLE,  # sanity check
            time=num_pv_timesteps,
        )
        hist_sat_trans_pv_attn_out = sat_trans_pv_attn_out[:, : pv_t0_idx + 1]

        pv_rnn_hist_enc_in = torch.concat((hist_pv, hist_sat_trans_pv_attn_out), dim=2)
        pv_rnn_hist_enc_in = pv_rnn_hist_enc_in.nan_to_num(0)
        pv_rnn_hist_enc_out, pv_rnn_hist_enc_hidden = self.pv_rnn_history_encoder(
            pv_rnn_hist_enc_in
        )

        # Now for the pv_rnn_future_decoder:
        future_sat_trans_pv_attn_out = sat_trans_pv_attn_out[:, pv_t0_idx + 1 :]
        future_sat_trans_pv_attn_out = future_sat_trans_pv_attn_out.nan_to_num(0)
        pv_rnn_fut_dec_out, _ = self.pv_rnn_future_decoder(
            future_sat_trans_pv_attn_out,
            pv_rnn_hist_enc_hidden,
        )

        # Concatenate the output from the encoder and decoder RNNs, and reshape
        # so each timestep and each PV system is a separate element into `time_transformer`.
        pv_rnn_out = torch.concat((pv_rnn_hist_enc_out, pv_rnn_fut_dec_out), dim=1)
        # rnn_out shape: (example n_pv_systems), time, d_model
        assert pv_rnn_out.isfinite().all()
        pv_rnn_out = einops.rearrange(
            pv_rnn_out,
            "(example n_pv_systems) time d_model -> example (time n_pv_systems) d_model",
            n_pv_systems=N_PV_SYSTEMS_PER_EXAMPLE,
            d_model=self.hidden_size,
            time=num_pv_timesteps,
        )
        return pv_rnn_out



from dataclasses import dataclass
from typing import Optional

import torch
from torch import nn


class TransformerEncoderLayerForVisualisingWeights(nn.TransformerEncoderLayer):
    def _sa_block(
        self,
        x: torch.Tensor,
        attn_mask: Optional[torch.Tensor],
        key_padding_mask: Optional[torch.Tensor],
    ) -> torch.Tensor:
        """Self-attention (sa) block."""
        # This version of the `_sa_block` method is functionally identical to the
        # one in nn.TransformerEncoderLayer.
        # The only change we make is to pass in all parameters into `self.self_attn` as
        # *positional* parameters (not keyword) so the `forward_pre_hook`
        # can access and modify them :). Because the `forward_pre_hook`
        # can only access positional params. It can't access keyword params.
        # Then we can use the `forward_pre_hook` to change `need_weights` to True.
        need_weights = False
        x = self.self_attn(x, x, x, key_padding_mask, need_weights, attn_mask)[0]
        return self.dropout1(x)


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
# for why we set `eq=False`
@dataclass(eq=False)
class MultiLayerTransformerEncoder(nn.Module):
    d_model: int
    num_heads: int = 8
    dropout: float = 0.0
    share_weights_across_latent_transformer_layers: bool = True
    num_latent_transformer_encoders: int = 4

    def __post_init__(self):
        super().__init__()

        # TransformerEncoderLayer is made up of self-attn and feedforward network.
        # This standard encoder layer is based on the paper “Attention Is All You Need”
        transformer_encoder_layer = TransformerEncoderLayerForVisualisingWeights(
            d_model=self.d_model,
            nhead=self.num_heads,
            dropout=self.dropout,
            batch_first=True,
            activation="gelu",
        )

        if self.share_weights_across_latent_transformer_layers:
            transformer_encoder_layers = [
                transformer_encoder_layer
            ] * self.num_latent_transformer_encoders
            self.transformer_encoder = nn.Sequential(*transformer_encoder_layers)
        else:
            self.transformer_encoder = nn.TransformerEncoder(
                encoder_layer=transformer_encoder_layer,
                num_layers=self.num_latent_transformer_encoders,
            )

    def forward(self, *args, **kwargs):
        return self.transformer_encoder(*args, **kwargs)


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
# for why we set `eq=False`
@dataclass(eq=False)
class Perceiver(nn.Module):
    """
    Input and output tensors are provided as shape (batch, seq, feature).

    Init args:
        num_heads: Number of parallel attention heads. Note that ``embed_dim`` will be split
            across ``num_heads`` (i.e. each head will have dimension ``embed_dim // num_heads``).
        dropout: Dropout probability on ``attn_output_weights``. Default: ``0.0`` (no dropout).
    """

    query_dim: int
    byte_array_dim: int
    num_heads: int = 8
    dropout: float = 0.0
    share_weights_across_latent_transformer_layers: bool = True
    num_latent_transformer_encoders: int = 4
    need_weights: bool = False

    def __post_init__(self):
        super().__init__()
        assert self.num_latent_transformer_encoders >= 1

        self.cross_attention = nn.MultiheadAttention(
            embed_dim=self.query_dim,
            num_heads=self.num_heads,
            dropout=self.dropout,
            batch_first=True,
            kdim=self.byte_array_dim,
            vdim=self.byte_array_dim,
        )

        self.latent_transformer = MultiLayerTransformerEncoder(
            d_model=self.query_dim,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

    def forward(
        self,
        query: torch.Tensor,
        byte_array: torch.Tensor,
        byte_array_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Args:
            query: Query embeddings of shape `(N, L, query_dim)`, where `N` is the batch size,
                and `L` is the number of query elements (the target sequence length).
                Queries are compared against key-value pairs to produce the output.
            byte_array: Key embeddings of shape `(N, S, byte_array_dim)`, where `N` is the
                batch size, and `S` is the number of byte_array elements
                (the source sequence length).
            byte_array_padding_mask: If specified, a mask of shape `(N, S)` indicating which
                elements within ``byte_array`` to ignore for the purpose of attention
                (i.e. treat as "padding"). Shape should be `(N, S)`. Binary and byte masks
                are supported. For a binary mask, a ``True`` value indicates that the corresponding
                ``byte_array`` value will be ignored for the purpose of attention. For a byte mask,
                a non-zero value indicates that the corresponding ``byte_array`` value
                will be ignored.

        Returns:
            **attn_output** - Attention outputs of shape `(N, L, query_dim)`, where `N` is the
                batch size, and `L` is the number of query elements (the target sequence length).
        """
        attn_output, attn_weights = self.cross_attention(
            query=query,
            key=byte_array,
            value=byte_array,
            need_weights=self.need_weights,
            key_padding_mask=byte_array_padding_mask,
        )
        return self.latent_transformer(attn_output)


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
# for why we set `eq=False`
@dataclass(eq=False)
class PerceiverIO(nn.Module):
    encoder_query_dim: int
    decoder_query_dim: int
    byte_array_dim: int
    num_encoder_heads: int
    num_decoder_heads: int
    dropout: float = 0.0
    share_weights_across_latent_transformer_layers: bool = True
    num_latent_transformer_encoders: int = 4
    num_cross_attends: int = 1

    def __post_init__(self):
        super().__init__()
        assert self.num_cross_attends >= 1

        self.perceiver_encoders = nn.ModuleList()
        for _ in range(self.num_cross_attends):
            self.perceiver_encoders.append(
                Perceiver(
                    query_dim=self.encoder_query_dim,
                    byte_array_dim=self.byte_array_dim,
                    num_heads=self.num_encoder_heads,
                    dropout=self.dropout,
                    share_weights_across_latent_transformer_layers=(
                        self.share_weights_across_latent_transformer_layers
                    ),
                    num_latent_transformer_encoders=self.num_latent_transformer_encoders,
                )
            )

        self.decoder = nn.MultiheadAttention(
            embed_dim=self.decoder_query_dim,
            num_heads=self.num_decoder_heads,
            dropout=self.dropout,
            batch_first=True,
            kdim=self.encoder_query_dim,
            vdim=self.encoder_query_dim,
        )

    def forward(
        self,
        encoder_query: torch.Tensor,
        byte_array: torch.Tensor,
        decoder_query: torch.Tensor,
        byte_array_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        # Encoder
        encoder_output = encoder_query
        for perceiver_encoder in self.perceiver_encoders:
            encoder_output = perceiver_encoder(
                query=encoder_output,
                byte_array=byte_array,
                byte_array_padding_mask=byte_array_padding_mask,
            )

        # Decoder
        decoder_output, decoder_weights = self.decoder(
            query=decoder_query,
            key=encoder_output,
            value=encoder_output,
            need_weights=False,
        )
        return decoder_output



import datetime
import logging
from dataclasses import dataclass
from typing import ClassVar

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr

from power_perceiver.consts import BatchKey, Location
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import NumpyBatch
from power_perceiver.load_raw.data_sources.raw_data_source import (
    RawDataSource,
    TimeseriesDataSource,
)
from power_perceiver.utils import check_path_exists, datetime64_to_float, select_time_periods

_log = logging.getLogger(__name__)


@dataclass(kw_only=True)
class RawGSPDataSource(
    # Surprisingly, Python's class hierarchy is defined right-to-left.
    # So the base class must go on the right.
    TimeseriesDataSource,
    RawDataSource,
):
    """
    Data source for GSP (Grid Supply Point) PV Data.

    Load GSP data directly from the intermediate GSP Zarr store.

    30 mins data is taken from 'PV Live' from https://www.solar.sheffield.ac.uk/pvlive/
    meta data is taken from ESO.  PV Live estimates the total PV power generation for each
    Grid Supply Point region.
    """

    gsp_pv_power_zarr_path: str
    gsp_id_to_region_id_filename: str
    sheffield_solar_region_path: str
    threshold_mw: int = 0

    sample_period_duration: ClassVar[datetime.timedelta] = datetime.timedelta(minutes=30)

    def __post_init__(self):  # noqa: D105
        self._sanity_check_args()
        RawDataSource.__post_init__(self)
        TimeseriesDataSource.__post_init__(self)
        # Load everything into RAM once (at init) rather than in each worker process.
        # This should be faster than loading from disk in every worker!
        self.load_everything_into_ram()
        self.empty_example = self._get_empty_example()

    def _sanity_check_args(self):  # noqa: D105
        check_path_exists(self.gsp_pv_power_zarr_path)
        check_path_exists(self.gsp_id_to_region_id_filename)
        check_path_exists(self.sheffield_solar_region_path)
        assert self.threshold_mw >= 0

    def load_everything_into_ram(self) -> None:
        """Open AND load GSP data into RAM."""
        gsp_id_to_shape = _get_gsp_id_to_shape(
            self.gsp_id_to_region_id_filename, self.sheffield_solar_region_path
        )
        self._gsp_id_to_shape = gsp_id_to_shape  # Save, mostly for plotting to check all is fine!

        # Load GSP generation xr.Dataset:
        gsp_pv_power_mw_ds = xr.open_dataset(self.gsp_pv_power_zarr_path, engine="zarr")
        gsp_pv_power_mw_ds = select_time_periods(
            xr_data=gsp_pv_power_mw_ds,
            time_periods=self.time_periods,
            dim_name="datetime_gmt",
        )
        gsp_pv_power_mw_ds = gsp_pv_power_mw_ds.load()

        # Ensure the centroids have the same GSP ID index as the GSP PV power:
        gsp_id_to_shape = gsp_id_to_shape.loc[gsp_pv_power_mw_ds.gsp_id]

        data_array = _put_gsp_data_into_an_xr_dataarray(
            gsp_pv_power_mw=gsp_pv_power_mw_ds.generation_mw.data.astype(np.float32),
            time_utc=gsp_pv_power_mw_ds.datetime_gmt.data,
            gsp_id=gsp_pv_power_mw_ds.gsp_id.data,
            # TODO: Try using `gsp_id_to_shape.geometry.envelope.centroid`. See issue #76.
            x_osgb=gsp_id_to_shape.geometry.centroid.x.astype(np.float32),
            y_osgb=gsp_id_to_shape.geometry.centroid.y.astype(np.float32),
            capacity_mwp=gsp_pv_power_mw_ds.installedcapacity_mwp.data.astype(np.float32),
            t0_idx=self.t0_idx,
        )

        del gsp_id_to_shape, gsp_pv_power_mw_ds

        # Select GSPs with sufficient installed PV capacity.
        # We take the `max` across time, because we don't want to drop GSPs which
        # do have installed PV capacity now, but didn't have PV power a while ago.
        # Although, in practice, swapping between `min` and `max` doesn't actually
        # seem to make any difference to the number of GSPs dropped :)
        gsp_id_mask = data_array.capacity_mwp.max(dim="time_utc") > self.threshold_mw
        data_array = data_array.isel(gsp_id=gsp_id_mask)
        data_array = data_array.dropna(dim="time_utc", how="all")

        _log.debug(
            f"There are {len(data_array.gsp_id)} GSPs left after we dropped"
            f" {(~gsp_id_mask).sum().data} GSPs because they had less than"
            f" {self.threshold_mw} MW of installed PV capacity."
            f" Total number of timesteps: {len(data_array.time_utc)}"
        )

        self._data_in_ram = data_array

    @property
    def num_gsps(self) -> int:
        return len(self.data_in_ram.gsp_id)

    def get_osgb_location_for_example(self) -> Location:
        """Get a single random geographical location."""
        random_gsp_idx = self.rng.integers(low=0, high=len(self.data_in_ram.gsp_id))
        return self.get_osgb_location_for_gsp_idx(random_gsp_idx)

    def get_osgb_location_for_gsp_idx(self, gsp_idx: int) -> Location:
        random_gsp = self.data_in_ram.isel(gsp_id=gsp_idx)
        return Location(x=random_gsp.x_osgb.item(), y=random_gsp.y_osgb.item())

    def _get_empty_example(self) -> xr.DataArray:
        """Get an empty example.

        The returned DataArray does not include an `example` dimension.
        """
        gsp_pv_power_mw = np.full(
            shape=(self.total_seq_length, 1), fill_value=np.NaN, dtype=np.float32
        )
        time_utc = np.full(shape=self.total_seq_length, fill_value=np.NaN, dtype=np.float32)
        time_utc = pd.DatetimeIndex(time_utc)
        gsp_meta = np.full(shape=1, fill_value=np.NaN, dtype=np.float32)

        return _put_gsp_data_into_an_xr_dataarray(
            gsp_pv_power_mw=gsp_pv_power_mw,
            time_utc=time_utc,
            gsp_id=gsp_meta,
            x_osgb=gsp_meta,
            y_osgb=gsp_meta,
            capacity_mwp=gsp_pv_power_mw,
            t0_idx=self.t0_idx,
        )

    def _get_spatial_slice(self, xr_data: xr.DataArray, center_osgb: Location) -> xr.DataArray:
        """Just return data for 1 GSP: The GSP whose centroid is (almost) equal to center_osgb."""
        mask = np.isclose(xr_data.x_osgb, center_osgb.x) & np.isclose(xr_data.y_osgb, center_osgb.y)
        return xr_data.isel(gsp_id=mask)

    def _post_process(self, xr_data: xr.DataArray) -> xr.DataArray:
        xr_data = xr_data / xr_data.capacity_mwp
        return xr_data

    @staticmethod
    def to_numpy(xr_data: xr.DataArray) -> NumpyBatch:
        """Return a single example in a `NumpyBatch`."""
        example: NumpyBatch = {}

        example[BatchKey.gsp] = xr_data.values
        example[BatchKey.gsp_t0_idx] = xr_data.attrs["t0_idx"]
        example[BatchKey.gsp_id] = xr_data.gsp_id.values
        example[BatchKey.gsp_capacity_mwp] = xr_data.isel(time_utc=0)["capacity_mwp"].values

        # Coordinates
        example[BatchKey.gsp_time_utc] = datetime64_to_float(xr_data["time_utc"].values)
        for batch_key, dataset_key in (
            (BatchKey.gsp_y_osgb, "y_osgb"),
            (BatchKey.gsp_x_osgb, "x_osgb"),
        ):
            values = xr_data[dataset_key].values
            # Expand dims so EncodeSpaceTime works!
            example[batch_key] = values  # np.expand_dims(values, axis=1)

        return example


def _get_gsp_id_to_shape(
    gsp_id_to_region_id_filename: str, sheffield_solar_region_path: str
) -> gpd.GeoDataFrame:
    # Load mapping from GSP ID to Sheffield Solar region ID:
    gsp_id_to_region_id = pd.read_csv(
        gsp_id_to_region_id_filename,
        usecols=["gsp_id", "region_id"],
        dtype={"gsp_id": np.int64, "region_id": np.int64},
    )

    # Load Sheffield Solar region shapes (which are already in OSGB36 CRS).
    ss_regions = gpd.read_file(sheffield_solar_region_path)

    # Merge, so we have a mapping from GSP ID to SS region shape:
    gsp_id_to_shape = (
        ss_regions.merge(gsp_id_to_region_id, left_on="RegionID", right_on="region_id")
        .set_index("gsp_id")[["geometry"]]
        .sort_index()
    )

    # Some GSPs are represented by multiple shapes. To find the correct centroid,
    # we need to find the spatial union of those regions, and then find the centroid
    # of those spatial unions. `dissolve(by="gsp_id")` groups by "gsp_id" and gets
    # the spatial union.
    return gsp_id_to_shape.dissolve(by="gsp_id")


def _put_gsp_data_into_an_xr_dataarray(
    gsp_pv_power_mw: np.ndarray,
    time_utc: np.ndarray,
    gsp_id: np.ndarray,
    x_osgb: np.ndarray,
    y_osgb: np.ndarray,
    capacity_mwp: np.ndarray,
    t0_idx: int,
) -> xr.DataArray:
    # Convert to xr.DataArray:
    data_array = xr.DataArray(
        gsp_pv_power_mw,
        coords=(("time_utc", time_utc), ("gsp_id", gsp_id)),
        name="gsp_pv_power_mw",
    )
    data_array = data_array.assign_coords(
        x_osgb=("gsp_id", x_osgb),
        y_osgb=("gsp_id", y_osgb),
        capacity_mwp=(("time_utc", "gsp_id"), capacity_mwp),
    )
    data_array.attrs["t0_idx"] = t0_idx
    return data_array



from dataclasses import dataclass
from typing import Optional, Union

import pandas as pd
import xarray as xr

from power_perceiver.time import set_new_sample_period_and_t0_idx_attrs


@dataclass
class PVPowerRollingWindow:
    """Compute rolling mean of PV power.

    The xr.Dataset is modified in-place, and is returned.

    Initialisation arguments: (taken from pandas.DataFrame.rolling docs)
        window: Size of the moving window.
            If an integer, the fixed number of observations used for each window.

            If an offset, the time period of each window. Each window will be a variable sized
            based on the observations included in the time-period. This is only valid for
            datetimelike indexes. To learn more about the offsets & frequency strings, please see:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases

            If a BaseIndexer subclass, the window boundaries based on the defined
            `get_window_bounds` method. Additional rolling keyword arguments,
            namely `min_periods` and `center` will be passed to `get_window_bounds`.

        min_periods: Minimum number of observations in window required to have a value;
            otherwise, result is `np.nan`.

            To avoid NaNs at the start and end of the timeseries, this should be <= ceil(window/2).

            For a window that is specified by an offset, `min_periods` will default to 1.

            For a window that is specified by an integer, `min_periods` will default to the size of
            the window.

        center: If False, set the window labels as the right edge of the window index.
            If True, set the window labels as the center of the window index.
    """

    window: Union[int, pd.tseries.offsets.DateOffset, pd.core.indexers.objects.BaseIndexer] = 3
    min_periods: Optional[int] = 2
    center: bool = True
    win_type: Optional[str] = None
    expect_dataset: bool = True

    def __call__(self, xr_data: Union[xr.Dataset, xr.DataArray]) -> Union[xr.Dataset, xr.DataArray]:
        if self.expect_dataset:
            data_to_resample = xr_data["power_w"]
        else:
            data_to_resample = xr_data

        resampled = data_to_resample.rolling(
            dim={"time_utc": self.window},
            min_periods=self.min_periods,
            center=self.center,
        ).mean()

        if self.expect_dataset:
            xr_data["power_w"] = resampled
            resampled = xr_data

        # Resampling removes the attributes, so put them back:
        for attr_name in ("t0_idx", "sample_period_duration"):
            resampled.attrs[attr_name] = xr_data.attrs[attr_name]

        return resampled


@dataclass
class PVDownsample:
    """Downsample PV.

    The xr.Dataset is modified in-place, and is returned.

    Initialisation arguments:
        freq:
        expect_dataset:
    """

    freq: str = "30T"
    expect_dataset: bool = True

    def __call__(self, xr_data: Union[xr.Dataset, xr.DataArray]) -> Union[xr.Dataset, xr.DataArray]:
        if self.expect_dataset:
            data_to_resample = xr_data["power_w"]
        else:
            data_to_resample = xr_data

        resampled = data_to_resample.resample(time_utc=self.freq, label="right").mean()

        if self.expect_dataset:
            xr_data["power_w"] = resampled
            resampled = xr_data

        # Resampling removes the attributes, so put them back:
        for attr_name in ("t0_idx", "sample_period_duration"):
            resampled.attrs[attr_name] = xr_data.attrs[attr_name]

        # Change the pv_t0_idx and the sample_period_duration attributes:
        resampled = set_new_sample_period_and_t0_idx_attrs(resampled, new_sample_period=self.freq)
        return resampled



# General imports
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Optional

import einops

# ML imports
import matplotlib.pyplot as plt
import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from pytorch_lightning.loggers import WandbLogger
from torch import nn
from torch.utils import data

from power_perceiver.analysis.plot_timeseries import LogTimeseriesPlots

# power_perceiver imports
from power_perceiver.analysis.plot_tsne import LogTSNEPlot
from power_perceiver.consts import BatchKey
from power_perceiver.load_prepared_batches.data_sources import GSP, PV, HRVSatellite, Sun
from power_perceiver.load_prepared_batches.prepared_dataset import PreparedDataset
from power_perceiver.np_batch_processor import EncodeSpaceTime, Topography
from power_perceiver.pytorch_modules.query_generator import GSPQueryGenerator, PVQueryGenerator
from power_perceiver.pytorch_modules.satellite_processor import HRVSatelliteProcessor
from power_perceiver.pytorch_modules.self_attention import MultiLayerTransformerEncoder
from power_perceiver.transforms.pv import PVPowerRollingWindow
from power_perceiver.transforms.satellite import PatchSatellite
from power_perceiver.xr_batch_processor import ReduceNumPVSystems, SelectPVSystemsNearCenterOfImage

plt.rcParams["figure.figsize"] = (18, 10)
plt.rcParams["figure.facecolor"] = "white"

DATA_PATH = Path(
    "/mnt/storage_ssd_4tb/data/ocf/solar_pv_nowcasting/nowcasting_dataset_pipeline/"
    "prepared_ML_training_data/v15/"
)
assert DATA_PATH.exists()

D_MODEL = 96


def get_dataloader(data_path: Path, tag: str) -> data.DataLoader:
    assert tag in ["train", "validation"]
    assert data_path.exists()

    dataset = PreparedDataset(
        data_path=data_path,
        data_loaders=[
            HRVSatellite(
                transforms=[
                    PatchSatellite(),
                ]
            ),
            PV(transforms=[PVPowerRollingWindow()]),
            Sun(),
            GSP(),
        ],
        xr_batch_processors=[
            SelectPVSystemsNearCenterOfImage(),
            ReduceNumPVSystems(requested_num_pv_systems=8),
        ],
        np_batch_processors=[
            EncodeSpaceTime(),
            Topography("/home/jack/europe_dem_2km_osgb.tif"),
        ],
    )

    def seed_rngs(worker_id: int):
        """Set different random seed per worker."""
        worker_info = torch.utils.data.get_worker_info()
        for xr_batch_processor in worker_info.dataset.xr_batch_processors:
            if getattr(xr_batch_processor, "rng", None):
                xr_batch_processor.rng = np.random.default_rng(seed=42 + worker_id)

    dataloader = data.DataLoader(
        dataset,
        batch_size=None,
        num_workers=16,
        pin_memory=True,
        worker_init_fn=seed_rngs,
    )

    return dataloader


train_dataloader = get_dataloader(DATA_PATH / "train", tag="train")
val_dataloader = get_dataloader(DATA_PATH / "test", tag="validation")


def maybe_pad_with_zeros(tensor: torch.Tensor, requested_dim: int) -> torch.Tensor:
    num_zeros_to_pad = requested_dim - tensor.shape[-1]
    assert num_zeros_to_pad >= 0, f"{requested_dim=}, {tensor.shape=}"
    if num_zeros_to_pad > 0:
        zero_padding_shape = tensor.shape[:2] + (num_zeros_to_pad,)
        zero_padding = torch.zeros(*zero_padding_shape, dtype=tensor.dtype, device=tensor.device)
        tensor = torch.concat((tensor, zero_padding), dim=2)
    return tensor


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class SatelliteTransformer(pl.LightningModule):
    """Infers a single timestep of PV power and GSP power at a time.

    Currently just uses HRV satellite imagery as the input. In the near future it could also
    use NWP temperature, wind speed & precipitation, and absolute geo position.
    """

    # Params for Perceiver
    # byte_array and query will be automatically padded with zeros to get to d_model.
    # Set d_model to be divisible by `num_heads`.
    d_model: int = D_MODEL
    pv_system_id_embedding_dim: int = 16
    num_heads: int = 12
    dropout: float = 0.0
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 4

    def __post_init__(self):
        super().__init__()
        self.hrvsatellite_processor = HRVSatelliteProcessor()

        NUM_PV_SYSTEMS = 1400
        NUM_GSPS = 360
        id_embedding = nn.Embedding(
            num_embeddings=NUM_PV_SYSTEMS + NUM_GSPS,
            embedding_dim=self.pv_system_id_embedding_dim,
        )

        self.pv_query_generator = PVQueryGenerator(pv_system_id_embedding=id_embedding)

        self.gsp_query_generator = GSPQueryGenerator(gsp_id_embedding=id_embedding)

        self.transformer_encoder = MultiLayerTransformerEncoder(
            d_model=self.d_model,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

        self.pv_output_module = nn.Sequential(
            nn.Linear(in_features=self.d_model, out_features=self.d_model),
            nn.GELU(),
            nn.Linear(in_features=self.d_model, out_features=1),
            nn.ReLU(),  # ReLU output guarantees that we can't predict negative PV power!
        )

    def forward(
        self,
        x: dict[BatchKey, torch.Tensor],
        start_idx_5_min: Optional[int] = None,
    ) -> dict[str, torch.Tensor]:
        if start_idx_5_min is None:
            assert self.training
            # Jitter the start_idx during training.
            device = x[BatchKey.pv].device
            start_idx_5_min = torch.randint(low=0, high=22, size=(1,), device=device)[0]

        start_idx_5_min_tensor = torch.tensor(start_idx_5_min)
        gsp_time_idx_30_min = (start_idx_5_min_tensor / 6).ceil().int() + 1

        byte_array = self.hrvsatellite_processor(x, start_idx_5_min=start_idx_5_min)
        pv_query = self.pv_query_generator(x, start_idx_5_min=start_idx_5_min)
        gsp_query = self.gsp_query_generator(x, time_idx_30_min=gsp_time_idx_30_min)

        # Pad with zeros if necessary to get up to self.d_model:
        byte_array = maybe_pad_with_zeros(byte_array, requested_dim=self.d_model)
        pv_query = maybe_pad_with_zeros(pv_query, requested_dim=self.d_model)
        gsp_query = maybe_pad_with_zeros(gsp_query, requested_dim=self.d_model)

        # Prepare the attention input and run through the transformer_encoder:
        attn_input = torch.concat((pv_query, gsp_query, byte_array), dim=1)
        attn_output = attn_input + self.transformer_encoder(attn_input)

        # Select the elements of the output which correspond to the query:
        gsp_start_idx = pv_query.shape[1]
        gsp_end_idx = gsp_start_idx + gsp_query.shape[1]
        pv_attn_out = attn_output[:, :gsp_start_idx]
        gsp_attn_out = attn_output[:, gsp_start_idx:gsp_end_idx]
        predicted_pv_power = self.pv_output_module(pv_attn_out)
        predicted_gsp_power = self.pv_output_module(gsp_attn_out)

        return {
            "pv_attn_out": pv_attn_out,  # shape: (example, n_pv_systems, d_model)
            "gsp_attn_out": gsp_attn_out,  # shape: (example, 1, d_model)
            "predicted_pv_power": predicted_pv_power,  # shape: (example, n_pv_systems=8, 1)
            "predicted_gsp_power": predicted_gsp_power,  # shape: (example, 1, 1)
            "start_idx_5_min": start_idx_5_min,
            "gsp_time_idx_30_min": gsp_time_idx_30_min,
        }


def get_multi_timestep_prediction(
    model: nn.Module,
    batch: dict[BatchKey, torch.Tensor],
    start_idxs_5_min: Iterable[int],
) -> dict[str, torch.Tensor]:
    predicted_pv_powers = []
    predicted_gsp_powers = []
    actual_gsp_powers = []
    gsp_times_utc = []
    pv_attn_outs = []
    gsp_attn_outs = []
    for start_idx_5_min in start_idxs_5_min:
        out = model.forward(batch, start_idx_5_min=start_idx_5_min)
        predicted_pv_powers.append(out["predicted_pv_power"])
        predicted_gsp_powers.append(out["predicted_gsp_power"][:, :, 0])
        gsp_time_idx_30_min = out["gsp_time_idx_30_min"]
        actual_gsp_power = batch[BatchKey.gsp][:, gsp_time_idx_30_min]  # Shape: (example)
        actual_gsp_power = actual_gsp_power.unsqueeze(1)  # Shape: (example, time=1)
        actual_gsp_powers.append(actual_gsp_power)
        gsp_time_utc = batch[BatchKey.gsp_time_utc][:, gsp_time_idx_30_min]
        gsp_time_utc = gsp_time_utc.unsqueeze(1)  # Shape: (example, time=1)
        gsp_times_utc.append(gsp_time_utc)
        pv_attn_outs.append(out["pv_attn_out"].unsqueeze(1))  # Shape: (example, 1, n_pv, d_mod)
        gsp_attn_outs.append(out["gsp_attn_out"])

    predicted_pv_power = torch.concat(predicted_pv_powers, dim=2)
    predicted_gsp_power = torch.concat(predicted_gsp_powers, dim=1)
    actual_gsp_power = torch.concat(actual_gsp_powers, dim=1)
    gsp_time_utc = torch.concat(gsp_times_utc, dim=1)
    pv_attn_out = torch.concat(pv_attn_outs, dim=1)
    gsp_attn_out = torch.concat(gsp_attn_outs, dim=1)

    del predicted_pv_powers, predicted_gsp_powers, actual_gsp_powers, gsp_times_utc
    del pv_attn_outs, gsp_attn_outs

    predicted_pv_power = einops.rearrange(
        predicted_pv_power,
        "example n_pv_systems time -> example time n_pv_systems",
        n_pv_systems=8,  # sanity check!
    )
    return dict(
        pv_attn_out=pv_attn_out,  # Shape: (example, time, n_pv_systems, d_model)
        gsp_attn_out=gsp_attn_out,  # Shape: (example, time, d_model)
        predicted_pv_power=predicted_pv_power,  # Shape: (example, time, n_pv_systems)
        predicted_gsp_power=predicted_gsp_power,  # Shape: (example, time)
        actual_gsp_power=actual_gsp_power,  # Shape: (example, time)
        gsp_time_utc=gsp_time_utc,  # Shape: (example, time)
    )


class TrainOrValidationMixIn:
    def _training_or_validation_step(
        self,
        batch: dict[BatchKey, torch.Tensor],
        batch_idx: int,
        tag: str,
        predicted_pv_power,
        predicted_gsp_power,
        actual_pv_power,
        actual_gsp_power,
        gsp_time_utc,
    ) -> dict[str, object]:
        """
        Args:
            batch: The training or validation batch.  A dictionary.
            tag: Either "train" or "validation"
            batch_idx: The index of the batch.
        """
        # Set PV power to zero for missing PV systems:
        actual_pv_power = torch.where(
            batch[BatchKey.pv_mask].unsqueeze(1),
            actual_pv_power,
            torch.tensor(0.0, dtype=actual_pv_power.dtype, device=actual_pv_power.device),
        )

        # PV power loss:
        # TODO: Fix this hack `[:, 2:]` when pre-training the "images to power" model!
        assert (
            predicted_pv_power.shape[1] == 4
        ), "TODO: Fix this hack `[:, 2:]` when pre-training the 'images to power' model!"
        pv_mse_loss = F.mse_loss(predicted_pv_power[:, 2:], actual_pv_power[:, 2:])
        pv_nmae_loss = F.l1_loss(predicted_pv_power[:, 2:], actual_pv_power[:, 2:])
        self.log(f"{tag}/pv_mse", pv_mse_loss)
        self.log(f"{tag}/pv_nmae", pv_nmae_loss)
        self.log(f"{tag}/mse", pv_mse_loss)  # To allow for each comparison to older models.

        # GSP power loss:
        gsp_mse_loss = F.mse_loss(predicted_gsp_power[:, 2:], actual_gsp_power[:, 2:])
        gsp_nmae_loss = F.l1_loss(predicted_gsp_power[:, 2:], actual_gsp_power[:, 2:])
        self.log(f"{tag}/gsp_mse", gsp_mse_loss)
        self.log(f"{tag}/gsp_nmae", gsp_nmae_loss)

        # Total MSE loss:
        total_mse_loss = pv_mse_loss + gsp_mse_loss
        self.log(f"{tag}/total_mse", total_mse_loss)

        return {
            "loss": total_mse_loss,
            "pv_mse_loss": pv_mse_loss,
            "gsp_mse_loss": gsp_mse_loss,
            "pv_nmae_loss": pv_nmae_loss,
            "gsp_nmae_loss": gsp_nmae_loss,
            "predicted_gsp_power": predicted_gsp_power,
            "actual_gsp_power": actual_gsp_power,
            "gsp_time_utc": gsp_time_utc,
            "actual_pv_power": actual_pv_power,
            "predicted_pv_power": predicted_pv_power,
        }


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class TrainSatelliteTransformer(pl.LightningModule, TrainOrValidationMixIn):
    def __post_init__(self):
        super().__init__()
        self.satellite_transformer = SatelliteTransformer.load_from_checkpoint(
            "/home/jack/dev/ocf/power_perceiver/experiments/model_checkpoints/018.13/model.ckpt"
        )

        # Do this at the end of __post_init__ to capture model topology to wandb:
        self.save_hyperparameters()

    def forward(
        self, x: dict[BatchKey, torch.Tensor], start_idx_5_min: Optional[int] = None
    ) -> dict[str, torch.Tensor]:
        return self.satellite_transformer(x=x, start_idx_5_min=start_idx_5_min)

    def training_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        """Train on a single (random) timestep per example."""
        out = self.forward(batch)
        start_idx_5_min = out["start_idx_5_min"]
        # The start_idx_5_min is the index into the time dimension of the start of the
        # sequence of satellite data.
        # Select PV data:
        # We want the PV data that is 30 minutes into the 45 min sequence of images:
        actual_pv_power = batch[BatchKey.pv][:, 6 + start_idx_5_min]
        actual_pv_power = actual_pv_power.unsqueeze(1)  # Shape: (example, time=1, pv_systems)
        predicted_pv_power = out["predicted_pv_power"]
        predicted_pv_power = einops.rearrange(
            predicted_pv_power,
            "example n_pv_systems 1 -> example 1 n_pv_systems",
            n_pv_systems=actual_pv_power.shape[2],  # sanity check!
        )
        # Select a single timestep of GSP data:
        gsp_time_idx_30_min = out["gsp_time_idx_30_min"]
        actual_gsp_power = batch[BatchKey.gsp][:, gsp_time_idx_30_min]  # Shape: (example)
        actual_gsp_power = actual_gsp_power.unsqueeze(1)  # Shape: (example, time=1)
        predicted_gsp_power = out["predicted_gsp_power"][:, :, 0]  # Shape: (example, time=1)
        gsp_time_utc = batch[BatchKey.gsp_time_utc][:, gsp_time_idx_30_min]
        gsp_time_utc = gsp_time_utc.unsqueeze(1)  # Shape: (example, time=1)
        return self._training_or_validation_step(
            batch=batch,
            batch_idx=batch_idx,
            tag="train",
            predicted_pv_power=predicted_pv_power,
            predicted_gsp_power=predicted_gsp_power,
            actual_pv_power=actual_pv_power,
            actual_gsp_power=actual_gsp_power,
            gsp_time_utc=gsp_time_utc,
        )

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        """Validate on multiple timesteps."""
        multi_timestep_prediction = get_multi_timestep_prediction(
            model=self, batch=batch, start_idxs_5_min=range(0, 22)
        )
        for key in ("pv_attn_out", "gsp_attn_out"):
            del multi_timestep_prediction[key]
        actual_pv_power = batch[BatchKey.pv][:, 6:-3]  # example, time, n_pv_systems
        return self._training_or_validation_step(
            batch=batch,
            batch_idx=batch_idx,
            tag="validation",
            actual_pv_power=actual_pv_power,
            **multi_timestep_prediction,
        )

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=5e-5)
        return optimizer


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class FullModel(pl.LightningModule, TrainOrValidationMixIn):
    d_model: int = D_MODEL  # Must be the same as for InferSingleTimestepOfPower
    pv_system_id_embedding_dim: int = 16
    num_heads: int = 12
    dropout: float = 0.1
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 2
    stop_gradients_upstream_of_time_transformer: bool = False
    train_infer_single_timestep_from_scratch: bool = True

    def __post_init__(self):
        super().__init__()
        if self.train_infer_single_timestep_from_scratch:
            assert not self.stop_gradients_upstream_of_time_transformer
            self.satellite_transformer = SatelliteTransformer()
        else:
            self.satellite_transformer = SatelliteTransformer.load_from_checkpoint(
                "/home/jack/dev/ocf/power_perceiver/experiments/model_checkpoints/018.13/model.ckpt"
            )

        assert self.d_model == self.satellite_transformer.d_model

        self.time_transformer = MultiLayerTransformerEncoder(
            d_model=self.d_model,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

        self.pv_output_module = nn.Sequential(
            nn.Linear(in_features=self.d_model, out_features=self.d_model),
            nn.GELU(),
            nn.Linear(in_features=self.d_model, out_features=1),
            nn.ReLU(),  # ReLU output guarantees that we can't predict negative PV power!
        )

        # Do this at the end of __post_init__ to capture model topology to wandb:
        self.save_hyperparameters()

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> dict[str, torch.Tensor]:
        START_IDX_5_MIN = 3
        END_IDX_5_MIN = 22
        STEP_5_MIN = 6  # Every half hour.
        multi_timestep_prediction = get_multi_timestep_prediction(
            model=self.satellite_transformer,
            batch=x,
            start_idxs_5_min=range(START_IDX_5_MIN, END_IDX_5_MIN, STEP_5_MIN),
        )

        pv_attn_out = multi_timestep_prediction["pv_attn_out"]
        n_timesteps, n_pv_systems = pv_attn_out.shape[1:3]
        pv_attn_out = einops.rearrange(
            pv_attn_out,
            "example time n_pv_systems d_model -> example (time n_pv_systems) d_model",
        )
        num_pv_elements = pv_attn_out.shape[1]
        gsp_attn_out = multi_timestep_prediction["gsp_attn_out"]

        # Get historical PV
        pv_query_generator = self.satellite_transformer.pv_query_generator
        historical_pv = []
        for start_idx_5_min in range(START_IDX_5_MIN, START_IDX_5_MIN + 13, STEP_5_MIN):
            hist_pv = pv_query_generator(
                x=x, start_idx_5_min=start_idx_5_min, num_timesteps_of_pv_power=STEP_5_MIN
            )
            hist_pv = maybe_pad_with_zeros(hist_pv, requested_dim=self.d_model)
            historical_pv.append(hist_pv)

        # Concatenate all the things we're going to feed into the "time transformer":
        time_attn_in = [pv_attn_out, gsp_attn_out] + historical_pv
        time_attn_in = torch.concat(time_attn_in, dim=1)
        if self.stop_gradients_upstream_of_time_transformer:
            # Detach so we don't train the previous model:
            time_attn_in = time_attn_in.detach()
        time_attn_out = self.time_transformer(time_attn_in)

        power_out = self.pv_output_module(time_attn_out)  # (example, total_num_elements, 1)

        # Reshape the PV power predictions
        predicted_pv_power = power_out[:, :num_pv_elements]
        predicted_pv_power = einops.rearrange(
            predicted_pv_power,
            "example (time n_pv_systems) 1 -> example time n_pv_systems",
            time=n_timesteps,
            n_pv_systems=n_pv_systems,
        )

        # Shape: (example time n_pv_systems)
        multi_timestep_prediction["predicted_pv_power"] = predicted_pv_power

        # GSP power. There's just 1 GSP.
        gsp_end_idx = num_pv_elements + n_timesteps
        predicted_gsp_power = power_out[:, num_pv_elements:gsp_end_idx, 0]
        # Shape: (example, time)
        multi_timestep_prediction["predicted_gsp_power"] = predicted_gsp_power

        return multi_timestep_prediction

    def training_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self.validation_step(batch, batch_idx)

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        """Validate on multiple timesteps."""
        multi_timestep_prediction = self(batch)
        for key in ("pv_attn_out", "gsp_attn_out"):
            del multi_timestep_prediction[key]
        actual_pv_power = batch[BatchKey.pv][:, 6:-3:6]  # example, time, n_pv_systems
        return self._training_or_validation_step(
            batch=batch,
            batch_idx=batch_idx,
            tag="train" if self.training else "validation",
            actual_pv_power=actual_pv_power,
            **multi_timestep_prediction,
        )

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=5e-5)
        return optimizer


model = FullModel()

wandb_logger = WandbLogger(
    name=(
        "020.00: 2 images straddling PV. d_model=96 num_timesteps=2"
        " start_idx_5_min_offset=4 interval=4"
    ),
    project="power_perceiver",
    entity="openclimatefix",
    log_model="all",
)

# log gradients, parameter histogram and model topology
# wandb_logger.watch(model, log="all")

trainer = pl.Trainer(
    gpus=[0],
    max_epochs=70,
    logger=wandb_logger,
    callbacks=[
        LogTimeseriesPlots(),
        LogTSNEPlot(query_generator_name="satellite_transformer.pv_query_generator"),
    ],
)

trainer.fit(
    model=model,
    train_dataloaders=train_dataloader,
    val_dataloaders=val_dataloader,
)



from dataclasses import dataclass

from power_perceiver.consts import BatchKey
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import NumpyBatch
from power_perceiver.utils import assert_num_dims


@dataclass
class DeleteForecastSatelliteImagery:
    """Delete imagery of the future.

    Useful when not training the U-Net, and we want to save GPU RAM.

    But we do want hrvsatellite_time_utc to continue out to 2 hours because
    downstream code relies on hrvsatellite_time_utc.
    """

    num_hist_sat_images: int

    def __call__(self, np_batch: NumpyBatch) -> NumpyBatch:
        # Shape: example, time, channels, y, x
        assert_num_dims(np_batch[BatchKey.hrvsatellite_actual], 5)
        np_batch[BatchKey.hrvsatellite_actual] = np_batch[BatchKey.hrvsatellite_actual][
            :, : self.num_hist_sat_images
        ]
        return np_batch



from dataclasses import dataclass

import numpy as np
import pandas as pd

from power_perceiver.consts import BatchKey
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import NumpyBatch
from power_perceiver.utils import datetime64_to_float, stack_np_examples_into_batch


@dataclass
class AlignGSPTo5Min:
    """Aligns GSP data to 5 min data.

    The GSP data isn't interpolated. Instead, for each 5_min_timestep, we take the GSP data at
    5_min_timestep.ceil("30T"). If that GSP timestep does not exist then NaNs will be used.
    """

    batch_key_for_5_min_datetimes: BatchKey = BatchKey.hrvsatellite_time_utc

    def __call__(self, np_batch: NumpyBatch) -> NumpyBatch:
        # Loop through each example and find the index into the GSP time dimension
        # of the GSP timestep corresponding to each 5 minute timestep:
        gsp_5_min_for_all_examples: list[NumpyBatch] = []
        n_examples = np_batch[BatchKey.gsp].shape[0]
        for example_i in range(n_examples):
            # Find the corresponding GSP 30 minute timestep for each 5 minute satellite timestep.
            # We do this by taking the `ceil("30T")` of each 5 minute satellite timestep.
            # Most of the code below is just converting to Pandas and back
            # so we can use `pd.DatetimeIndex.ceil` on each datetime:
            time_5_min = np_batch[self.batch_key_for_5_min_datetimes][example_i]
            time_5_min_dt_index = pd.to_datetime(time_5_min, unit="s")
            time_30_min_every_5_min_dt_index = time_5_min_dt_index.ceil("30T")
            time_30_min_every_5_min = datetime64_to_float(time_30_min_every_5_min_dt_index.values)

            # Now, find the index into the original 30-minute GSP data for each 5-min timestep:
            gsp_30_min_time = np_batch[BatchKey.gsp_time_utc][example_i]
            idx_into_gsp = np.searchsorted(gsp_30_min_time, time_30_min_every_5_min)

            gsp_5_min_example: NumpyBatch = {}
            for batch_key in (BatchKey.gsp, BatchKey.gsp_time_utc):
                new_batch_key_name = batch_key.name.replace("gsp", "gsp_5_min")
                new_batch_key = BatchKey[new_batch_key_name]
                gsp_5_min_example[new_batch_key] = np_batch[batch_key][example_i, idx_into_gsp]

            gsp_5_min_for_all_examples.append(gsp_5_min_example)

        # Stack the individual examples back into a batch of examples:
        new_np_batch = stack_np_examples_into_batch(gsp_5_min_for_all_examples)
        np_batch.update(new_np_batch)

        # Copy over the t0_idx scalar:
        batch_key_name_for_5_min_t0_idx = self.batch_key_for_5_min_datetimes.name.replace(
            "time_utc", "t0_idx"
        )
        batch_key_for_5_min_t0_idx = BatchKey[batch_key_name_for_5_min_t0_idx]
        np_batch[BatchKey.gsp_5_min_t0_idx] = np_batch[batch_key_for_5_min_t0_idx]

        return np_batch



import pytest

from power_perceiver.production.batch import get_batch


def test_pv(pv):
    pass


def test_gsp(gsp):
    pass


def test_nwp(nwp):
    pass


def test_hrv_satellite(hrv_satellite):
    pass


def test_pv_check_negative(pv):
    with pytest.raises(Exception):
        pv_x_osgb = pv.pv_x_osgb
        pv_x_osgb[0, 0] = -1
        pv.pv_x_osgb = pv_x_osgb


def test_pv_check_wrong_shape(pv):
    with pytest.raises(Exception):
        pv_x_osgb = pv.pv_x_osgb
        pv_x_osgb = pv_x_osgb[0]
        pv.pv_x_osgb = pv_x_osgb


def test_make_batch(pv, nwp, gsp, hrv_satellite):
    _ = get_batch(pv=pv, nwp=nwp, gsp=gsp, hrvsatelle=hrv_satellite)



import datetime
import logging
from dataclasses import dataclass
from numbers import Number
from typing import Callable, Iterator, Optional, Sequence

import numpy as np
import pandas as pd
import torch

from power_perceiver.consts import BatchKey, Location
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import (
    NumpyBatch,
    XarrayBatch,
)
from power_perceiver.load_raw.data_sources.raw_data_source import (
    RawDataSource,
    TimeseriesDataSource,
)
from power_perceiver.time import (
    intersection_of_multiple_dataframes_of_periods,
    time_periods_to_datetime_index,
)
from power_perceiver.utils import (
    sample_row_and_drop_row_from_df,
    set_fsspec_for_multiprocess,
    stack_np_examples_into_batch,
)

_log = logging.getLogger(__name__)


@dataclass
class RawDataset(torch.utils.data.IterableDataset):
    """Dataset for loading data from intermediate data (not pre-prepared batches).

    Initialisation arguments:
        data_source_combos: A dict where the keys are strings (short, arbitrary names
            identifying each combination of DataSources), and the values are tuples or lists of
            instantiated `DataSource` objects. The first data source
            will randomly select the geographical location of each example.

            If a DataSource appears more than once, then you must use different instances for each
            combo (e.g. by using `deepcopy`). Using different instances
            allows each data source combo to load time periods appropriate for that combo.

            For example:
                data_source_kwargs = dict(start_date=start_date, end_date=end_date)
                sat = RawSatelliteDataSource(**data_source_kwargs)
                gsp = RawGSPDataSource(**data_source_kwargs)
                data_source_combos = dict(
                     sat_only=(sat,),
                     gsp_and_sat=(gsp, deepcopy(sat)))

        probability_of_each_combo: A dict where the key is the name of each
            combination of DataSources, and the key is the probability of loading that
            combination of DataSources. Probabilities must sum to 1.
            Optional. If `None` then will use equal probabilities.
        min_duration_to_load_per_epoch: If this an int, then this will be (roughly) the total
            number of hours of contiguous time periods that will be loaded into RAM
            at the start of each epoch. Some DataSources (such as PV) can load the entire
            dataset into RAM at initialization. If using only DataSources which can load
            everything into RAM, then set `min_duration_to_load_per_epoch` to None to avoid
            computing the subset of data to load into RAM at the start of each epoch.
            If `min_duration_to_load_per_epoch` is not `None`, then `ds_combo_for_subsetting` must
            not be `None`.
        load_subset_every_epoch: Set to False for use as a validation dataset.
        n_batches_per_epoch:
        n_examples_per_batch:
        np_batch_processors: Functions which takes a NumpyBatch, and processes
            *across* modalities, and returns the processed NumpyBatch. Note that numpy
            processing *within* a modality should be done in `DataSource.to_numpy`.
        t0_freq: The temporal frequency at which to convert time periods to datetimes.

    Attributes:
        _all_t0_periods_per_combo: dict[str, pd.DataFrame] where each row of each DataFrame
            represents a single time period. Each pd.DataFrame has two columns:
            `start_dt` and `end_dt` (where 'dt' is short for 'datetime').
            The keys of the dict are the same short names of the combination of DataSources.
            This is the total t0 time periods available, before any subsetting to load into RAM.
        _t0_datetimes_per_combo_for_epoch: dict[str, pd.DatetimeIndex]
    """

    data_source_combos: dict[str, Sequence[RawDataSource]]
    probability_of_each_combo: Optional[dict[str, Number]] = None
    min_duration_to_load_per_epoch: Optional[datetime.timedelta] = datetime.timedelta(hours=48 * 12)
    load_subset_every_epoch: bool = True
    n_batches_per_epoch: int = 1024
    n_examples_per_batch: int = 32
    np_batch_processors: Optional[Sequence[Callable]] = None
    t0_freq: str = "5T"

    def __post_init__(self):  # noqa: D105
        self._sanity_check_args()
        super().__init__()
        self._all_t0_periods_per_combo: dict[str, pd.DataFrame] = {}
        self._t0_datetimes_per_combo_for_epoch: dict[str, pd.DatetimeIndex] = {}

    def _sanity_check_args(self):  # noqa: D105
        if self.probability_of_each_combo is not None:
            assert self.data_source_combos.keys() == self.probability_of_each_combo.keys()
        if self.min_duration_to_load_per_epoch is not None:
            assert self.min_duration_to_load_per_epoch > pd.Timedelta(self.t0_freq)
        assert self.n_batches_per_epoch > 0

    def per_worker_init(self, worker_id: int = 0, seed: int = 42) -> None:
        """Called by worker_init_fn on each copy of this dataset after the
        worker process has been spawned."""
        set_fsspec_for_multiprocess()
        self.worker_id = worker_id
        # Each worker must have a different seed for its random number generator.
        # Otherwise all the workers will output exactly the same data!
        # The worker ID will be different for each worker process for each GPU.
        _log.info(f"{worker_id=} has random number generator {seed=:,d}")
        self.rng = np.random.default_rng(seed=seed)

        for data_source in self._unique_data_sources:
            data_source.per_worker_init(worker_id=worker_id, seed=seed)

        self._get_intersection_of_all_t0_periods_per_combo()

    def _get_intersection_of_all_t0_periods_per_combo(self) -> None:
        """Get intersection of all contig t0 periods per combo, before subsetting.

        Sets `self._all_t0_periods_per_combo`
        """
        self._all_t0_periods_per_combo = {}
        for combo_name, data_sources in self.data_source_combos.items():
            t0_periods_for_combo: list[pd.DataFrame] = []
            for data_source in data_sources:
                if isinstance(data_source, TimeseriesDataSource):
                    t0_periods_for_combo.append(data_source.get_contiguous_t0_time_periods())
            self._all_t0_periods_per_combo[
                combo_name
            ] = intersection_of_multiple_dataframes_of_periods(t0_periods_for_combo)

    def __iter__(self) -> Iterator[NumpyBatch]:  # noqa: D105
        self._epoch_start()
        for _ in range(self.n_batches_per_epoch):
            yield self._get_np_batch()

    def _epoch_start(self) -> None:
        if self.load_subset_every_epoch or not self._t0_datetimes_per_combo_for_epoch:
            self._set_t0_datetimes_per_combo_for_epoch_and_maybe_load_subset_into_ram()

    def _set_t0_datetimes_per_combo_for_epoch_and_maybe_load_subset_into_ram(self):
        # Compute subset of contiguous t0 time periods for this epoch, and ask each unique
        # data source that needs to load a subset into RAM to do so:
        data_sources_which_have_loaded = []
        for combo_name, data_sources in self.data_source_combos.items():
            data_sources_which_need_to_load = [
                ds for ds in data_sources if ds.needs_to_load_subset_into_ram
            ]
            if data_sources_which_need_to_load:
                t0_periods_for_combo_for_epoch = self._subset_t0_periods(combo_name)

                # Load into RAM, if necessary:
                for data_source in data_sources_which_need_to_load:
                    # Check we haven't already loaded this data_source in another combo:
                    if not any([data_source is ds for ds in data_sources_which_have_loaded]):
                        data_source.load_subset_into_ram(t0_periods_for_combo_for_epoch)
            else:
                t0_periods_for_combo_for_epoch = self._all_t0_periods_per_combo[combo_name]

            self._t0_datetimes_per_combo_for_epoch[combo_name] = time_periods_to_datetime_index(
                time_periods=t0_periods_for_combo_for_epoch, freq=self.t0_freq
            )

    def _subset_t0_periods(self, combo_name: str) -> pd.DataFrame:
        """Pick a random selection of contiguous time periods for the upcoming epoch.

        The main use-case for this is for `RawSatelliteDataSource` which needs to load a subset of
        data into RAM at the start of each epoch.
        """
        _log.info(f"Selecting a subset of time periods to load into RAM for {combo_name=}.")
        assert (
            self._all_t0_periods_per_combo
        ), "self._all_t0_periods_per_combo is empty! Have you forgotten to call per_worker_init?"
        all_t0_periods_for_combo = self._all_t0_periods_per_combo[combo_name]

        # Select random periods. We use a `while` loop instead of just doing
        # `rng.choice(all_t0_periods_for_combo, size=n)` because using `rng.choice`
        # wouldn't take into consideration that some periods are longer than others,
        # and we want to load roughly the same duration of data per epoch.
        random_t0_periods: list[pd.Series[str, pd.Timestamp]] = []
        total_duration_of_periods = pd.Timedelta(0)
        while total_duration_of_periods < self.min_duration_to_load_per_epoch:
            period, all_t0_periods_for_combo = sample_row_and_drop_row_from_df(
                all_t0_periods_for_combo, rng=self.rng
            )
            random_t0_periods.append(period)
            period_duration = period.end_dt - period.start_dt
            total_duration_of_periods += period_duration
            if all_t0_periods_for_combo.empty:
                break

        random_t0_periods = pd.DataFrame(random_t0_periods).sort_values("start_dt")
        _log.info(
            f"{self.worker_id=}. Selected {len(random_t0_periods):,d} random periods,"
            f" with total duration = {total_duration_of_periods},"
            f"from {random_t0_periods.iloc[0].start_dt} to {random_t0_periods.iloc[-1].end_dt}"
        )
        return random_t0_periods

    def _get_np_batch(self) -> NumpyBatch:
        np_examples = [self._get_np_example() for _ in range(self.n_examples_per_batch)]
        np_batch = stack_np_examples_into_batch(np_examples)
        return self._process_np_batch(np_batch)

    def _get_np_example(self) -> NumpyBatch:
        xr_example = self._get_xr_example()
        return self._xarray_to_numpy_example(xr_example)

    def _get_xr_example(self) -> XarrayBatch:
        chosen_combo_name = self._choose_combo_name()
        t0_datetime_utc = self._choose_t0_datetime(chosen_combo_name)
        location_osgb = self._choose_osgb_location(chosen_combo_name)

        try:
            xr_example = self._get_specific_xr_example(
                chosen_combo_name=chosen_combo_name,
                t0_datetime_utc=t0_datetime_utc,
                location_osgb=location_osgb,
            )
        except Exception as e:
            raise e.__class__(f"{chosen_combo_name=}; {t0_datetime_utc=}; {location_osgb=}") from e

        return xr_example

    def _choose_combo_name(self) -> str:
        """Pick a random ds_combo_name using the probabilities."""
        data_source_combo_names = list(self.data_source_combos.keys())
        if self.probability_of_each_combo is None:
            prob_of_each_combo = None
        else:
            # Need to ensure the probabilities are in the same order as the combos!
            prob_of_each_combo = [
                self.probability_of_each_combo[combo_name] for combo_name in data_source_combo_names
            ]
        return self.rng.choice(data_source_combo_names, p=prob_of_each_combo)

    def _choose_t0_datetime(self, chosen_combo_name: str) -> datetime.datetime:
        return self.rng.choice(self._t0_datetimes_per_combo_for_epoch[chosen_combo_name])

    def _choose_osgb_location(self, chosen_combo_name: str) -> Location:
        data_source_combo = self.data_source_combos[chosen_combo_name]
        data_source_which_selects_location = data_source_combo[0]
        return data_source_which_selects_location.get_osgb_location_for_example()

    def _get_specific_xr_example(
        self,
        chosen_combo_name: str,
        t0_datetime_utc: datetime.datetime,
        location_osgb: Location,
    ) -> XarrayBatch:
        # Loop through each data source in the combo:
        chosen_data_source_combo = self.data_source_combos[chosen_combo_name]
        xr_example: XarrayBatch = {}
        for data_source in self._unique_data_sources:
            if data_source in chosen_data_source_combo:
                xr_example[data_source.__class__] = data_source.get_example(
                    t0_datetime_utc, location_osgb
                )
            else:
                xr_example[data_source.__class__] = data_source.empty_example

        return xr_example

    def _xarray_to_numpy_example(self, xr_example: XarrayBatch) -> NumpyBatch:
        """Convert from xarray Datasets to numpy."""
        np_example: NumpyBatch = {}
        for data_loader_class, xr_dataset in xr_example.items():
            if data_loader_class == BatchKey.requested_timesteps:
                # `ReduceNumTimesteps` introduces a `requested_timesteps` key,
                # whose value is a np.ndarray.
                requested_timesteps = xr_dataset
                np_example[BatchKey.requested_timesteps] = requested_timesteps
            else:
                np_data_for_data_source = data_loader_class.to_numpy(xr_dataset)
                np_example.update(np_data_for_data_source)
        return np_example

    def _process_np_batch(self, np_batch: NumpyBatch) -> NumpyBatch:
        """If necessary, do any processing which needs to be done across modalities,
        on the NumpyBatch."""
        if self.np_batch_processors:
            for np_batch_processor in self.np_batch_processors:
                np_batch = np_batch_processor(np_batch)
        return np_batch

    @property
    def _unique_data_sources(self) -> list[RawDataSource]:
        # We can't use `set()` or `np.unique()` on a list of `RawDataSource` objects.
        unique_data_sources: list[RawDataSource] = []
        for tuple_of_data_sources in self.data_source_combos.values():
            for data_source in tuple_of_data_sources:
                # We need to use `is` to check if data_source *is* the exact same instance
                # as a RawDataSource already in `unique_data_sources`.
                # Using `if data_source in unique_data_sources` doesn't work because
                # that just uses the equality operator, which returns True if two
                # objects are identical but not necessarily the same instance.
                if not any([data_source is ds for ds in unique_data_sources]):
                    unique_data_sources.append(data_source)
        return unique_data_sources



# General imports
import logging
import socket
from dataclasses import dataclass, field
from pathlib import Path

import einops

# ML imports
import matplotlib.pyplot as plt
import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from pytorch_lightning.loggers import WandbLogger
from pytorch_msssim import ms_ssim
from torch import nn

# from power_perceiver.analysis.plot_satellite import LogSatellitePlots
from power_perceiver.analysis.plot_timeseries import LogTimeseriesPlots
from power_perceiver.analysis.plot_tsne import LogTSNEPlot

# power_perceiver imports
from power_perceiver.consts import (
    T0_IDX_30_MIN,
    X_OSGB_MEAN,
    X_OSGB_STD,
    Y_OSGB_MEAN,
    Y_OSGB_STD,
    BatchKey,
)
from power_perceiver.load_prepared_batches.data_sources import HRVSatellite
from power_perceiver.load_prepared_batches.data_sources.gsp import GSP
from power_perceiver.load_prepared_batches.data_sources.pv import PV
from power_perceiver.load_prepared_batches.data_sources.satellite import SAT_MEAN, SAT_STD
from power_perceiver.load_prepared_batches.data_sources.sun import Sun
from power_perceiver.load_prepared_batches.prepared_dataset import PreparedDataset
from power_perceiver.np_batch_processor.align_gsp_to_5_min import GSP5Min
from power_perceiver.np_batch_processor.encode_space_time import EncodeSpaceTime
from power_perceiver.np_batch_processor.sun_position import SunPosition
from power_perceiver.np_batch_processor.topography import Topography
from power_perceiver.pytorch_modules.query_generator import GSPQueryGenerator, PVQueryGenerator
from power_perceiver.pytorch_modules.satellite_predictor import XResUNet
from power_perceiver.pytorch_modules.satellite_processor import HRVSatelliteProcessor
from power_perceiver.pytorch_modules.self_attention import MultiLayerTransformerEncoder
from power_perceiver.transforms.pv import PVPowerRollingWindow
from power_perceiver.xr_batch_processor import (
    AlignGSPTo5Min,
    ReduceNumPVSystems,
    SelectPVSystemsNearCenterOfImage,
)
from power_perceiver.xr_batch_processor.reduce_num_timesteps import ReduceNumTimesteps

logging.basicConfig()
_log = logging.getLogger("power_perceiver")
_log.setLevel(logging.DEBUG)

plt.rcParams["figure.figsize"] = (18, 10)
plt.rcParams["figure.facecolor"] = "white"

# SatellitePredictor options
NUM_HIST_SAT_IMAGES = 7  # v15 pre-prepared batches use 7
NUM_FUTURE_SAT_IMAGES = 24  # v15 pre-prepared batches use 24
IMAGE_SIZE_PIXELS = 64  # v15 pre-prepared batches use 64
USE_TOPOGRAPHY = True
USE_SUN_POSITION = True

# PowerPerceiver options
D_MODEL = 128
N_HEADS = 16
T0_IDX_5_MIN_TRAINING = T0_IDX_5_MIN_VALIDATION = NUM_HIST_SAT_IMAGES - 1

if socket.gethostname() == "donatello":
    DATA_PATH = Path(
        "/mnt/storage_ssd_4tb/data/ocf/solar_pv_nowcasting/nowcasting_dataset_pipeline/"
        "prepared_ML_training_data/v15"
    )
else:
    # On Google Cloud VM:
    DATA_PATH = Path("/home/jack/data/v15")

assert DATA_PATH.exists()


torch.manual_seed(42)


def get_dataloader(data_path: Path, tag: str) -> torch.utils.data.DataLoader:
    assert tag in ["train", "validation"]
    assert data_path.exists()

    xr_batch_processors = [
        SelectPVSystemsNearCenterOfImage(),
        ReduceNumPVSystems(requested_num_pv_systems=8),
        AlignGSPTo5Min(),
    ]
    if tag == "train":
        xr_batch_processors.append(
            ReduceNumTimesteps(
                keys=(HRVSatellite, PV, GSP5Min, Sun),
                # When using UNet, every example must include the full complement
                # of historical satellite images.
                num_requested_history_timesteps=NUM_HIST_SAT_IMAGES,
                num_history_timesteps_available=NUM_HIST_SAT_IMAGES,
            )
        )

    np_batch_processors = [EncodeSpaceTime()]
    if USE_SUN_POSITION:
        np_batch_processors.append(SunPosition(t0_timestep_idx=NUM_HIST_SAT_IMAGES - 1))
    if USE_TOPOGRAPHY:
        np_batch_processors.append(Topography("/home/jack/europe_dem_2km_osgb.tif"))

    dataset = PreparedDataset(
        data_path=data_path,
        data_loaders=[
            # Note that we're NOT using `PatchSatellite`. Instead we use `einops` to
            # patch the satellite data. The U-Net doesn't work with `PatchSatellite`,
            # and it's easier to use `einops` to patch the satellite and downsample the
            # position encodings in PyTorch land.
            HRVSatellite(),
            PV(transforms=[PVPowerRollingWindow()]),
            Sun(),
            GSP(),
        ],
        xr_batch_processors=xr_batch_processors,
        np_batch_processors=np_batch_processors,
    )

    def seed_rngs(worker_id: int):
        """Set different random seed per worker."""
        worker_info = torch.utils.data.get_worker_info()
        for xr_batch_processor in worker_info.dataset.xr_batch_processors:
            if getattr(xr_batch_processor, "rng", None):
                xr_batch_processor.rng = np.random.default_rng(seed=42 + worker_id)

    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_size=None,
        num_workers=12,
        pin_memory=True,
        worker_init_fn=seed_rngs,
    )

    return dataloader


train_dataloader = get_dataloader(DATA_PATH / "train", tag="train")
val_dataloader = get_dataloader(DATA_PATH / "test", tag="validation")


# ---------------------------------- SatellitePredictor ----------------------------------


def get_osgb_coords_for_coord_conv(batch: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
    """Returns tensor of shape (example, 2, y, x)."""
    y_osgb = batch[BatchKey.hrvsatellite_y_osgb]
    x_osgb = batch[BatchKey.hrvsatellite_x_osgb]

    # Normalise:
    y_osgb = (y_osgb - Y_OSGB_MEAN) / Y_OSGB_STD
    x_osgb = (x_osgb - X_OSGB_MEAN) / X_OSGB_STD

    # Concat:
    return torch.stack((y_osgb, x_osgb), dim=1)


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class SatellitePredictor(pl.LightningModule):
    use_coord_conv: bool = False
    crop: bool = False
    optimizer_class: torch.optim.Optimizer = torch.optim.Adam
    optimizer_kwargs: dict = field(
        # lambda trick from https://stackoverflow.com/a/52064202/732596
        default_factory=lambda: dict(lr=1e-4)
    )
    use_topography: bool = USE_TOPOGRAPHY
    use_sun_position: bool = USE_SUN_POSITION

    # kwargs to fastai DynamicUnet. See this page for details:
    # https://fastai1.fast.ai/vision.models.unet.html#DynamicUnet
    pretrained: bool = False
    blur_final: bool = True  # Blur final layer. fastai default is True.
    self_attention: bool = True  # Use SA layer at the third block before the end.
    last_cross: bool = True  # Use a cross-connection with the direct input of the model.
    bottle: bool = False  # Bottleneck the last skip connection.

    def __post_init__(self):
        super().__init__()

        self.satellite_predictor = XResUNet(
            img_size=(IMAGE_SIZE_PIXELS, IMAGE_SIZE_PIXELS),
            n_in=(
                NUM_HIST_SAT_IMAGES
                + (2 if self.use_coord_conv else 0)
                + (1 if self.use_topography else 0)
                + (2 if self.use_sun_position else 0)
            ),
            n_out=NUM_FUTURE_SAT_IMAGES,
            pretrained=self.pretrained,
            blur_final=self.blur_final,
            self_attention=self.self_attention,
            last_cross=self.last_cross,
            bottle=self.bottle,
        )

        # Do this at the end of __post_init__ to capture model topology to wandb:
        self.save_hyperparameters()

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
        data = x[BatchKey.hrvsatellite_actual][
            :, :NUM_HIST_SAT_IMAGES, 0
        ]  # Shape: (example, time, y, x)
        height, width = data.shape[2:]
        assert height == IMAGE_SIZE_PIXELS, f"{height=}"
        assert width == IMAGE_SIZE_PIXELS, f"{width=}"
        if self.use_coord_conv:
            osgb_coords = get_osgb_coords_for_coord_conv(x)
            data = torch.concat((data, osgb_coords), dim=1)

        if self.use_topography:
            surface_height = x[BatchKey.hrvsatellite_surface_height]
            surface_height = surface_height.unsqueeze(1)  # Add channel dim
            data = torch.concat((data, surface_height), dim=1)

        if self.use_sun_position:
            azimuth = x[BatchKey.solar_azimuth_at_t0]
            elevation = x[BatchKey.solar_elevation_at_t0]
            sun_pos = torch.stack((azimuth, elevation), dim=1)  # Shape: (example, 2)
            del azimuth, elevation
            # Repeat over y and x:
            sun_pos = einops.repeat(sun_pos, "example chan -> example chan y x", y=height, x=width)
            data = torch.concat((data, sun_pos), dim=1)

        predicted_sat = self.satellite_predictor(data)
        return predicted_sat  # Shape: example, time, y, x

    def training_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self.validation_step(batch, batch_idx)

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        tag = "train" if self.training else "validation"
        network_out = self(batch)
        predicted_sat = network_out
        actual_sat = batch[BatchKey.hrvsatellite_actual][:, NUM_HIST_SAT_IMAGES:, 0]
        sat_mse_loss = F.mse_loss(predicted_sat, actual_sat)
        self.log(f"{tag}/sat_mse", sat_mse_loss)

        # MS-SSIM. Requires images to be de-normalised:
        actual_sat_denorm = (actual_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        predicted_sat_denorm = (predicted_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        ms_ssim_loss = 1 - ms_ssim(
            predicted_sat_denorm,
            actual_sat_denorm,
            data_range=1023.0,
            size_average=True,  # Return a scalar.
            win_size=3,  # ClimateHack folks used win_size=3.
        )
        self.log(f"{tag}/ms_ssim", ms_ssim_loss)
        self.log(f"{tag}/ms_ssim+sat_mse", ms_ssim_loss + sat_mse_loss)

        if self.crop:
            # Loss on 33x33 central crop:
            # The image has to be larger than 32x32 otherwise ms-ssim complains:
            # "Image size should be larger than 32 due to the 4 downsamplings in ms-ssim"
            CROP = 15
            sat_mse_loss_crop = F.mse_loss(
                predicted_sat[:, :, CROP:-CROP, CROP:-CROP],
                actual_sat[:, :, CROP:-CROP, CROP:-CROP],
            )
            self.log(f"{tag}/sat_mse_crop", sat_mse_loss_crop)
            ms_ssim_loss_crop = 1 - ms_ssim(
                predicted_sat_denorm[:, :, CROP:-CROP, CROP:-CROP],
                actual_sat_denorm[:, :, CROP:-CROP, CROP:-CROP],
                data_range=1023.0,
                size_average=True,  # Return a scalar.
                win_size=3,  # ClimateHack folks used win_size=3.
            )
            self.log(f"{tag}/ms_ssim_crop", ms_ssim_loss_crop)
            self.log(f"{tag}/ms_ssim_crop+sat_mse_crop", ms_ssim_loss_crop + sat_mse_loss_crop)
            loss = ms_ssim_loss_crop + sat_mse_loss_crop
        else:
            loss = ms_ssim_loss + sat_mse_loss

        return dict(
            loss=loss,
            predicted_sat=predicted_sat,
            actual_sat=actual_sat,
        )

    def configure_optimizers(self):
        optimizer = self.optimizer_class(self.parameters(), **self.optimizer_kwargs)

        def _lr_lambda(epoch):
            return 50 / (epoch + 50)

        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, _lr_lambda, verbose=True)
        return [optimizer], [scheduler]


# ---------------------------------- SatelliteTransformer ----------------------------------


def maybe_pad_with_zeros(tensor: torch.Tensor, requested_dim: int) -> torch.Tensor:
    num_zeros_to_pad = requested_dim - tensor.shape[-1]
    assert num_zeros_to_pad >= 0, f"{requested_dim=}, {tensor.shape=}"
    if num_zeros_to_pad > 0:
        zero_padding_shape = tensor.shape[:2] + (num_zeros_to_pad,)
        zero_padding = torch.zeros(*zero_padding_shape, dtype=tensor.dtype, device=tensor.device)
        tensor = torch.concat((tensor, zero_padding), dim=2)
    return tensor


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class SatelliteTransformer(nn.Module):
    """Infers a single timestep of PV power and GSP power at a time.

    Currently just uses HRV satellite imagery as the input. In the near future it could also
    use NWP temperature, wind speed & precipitation, and absolute geo position.
    """

    # Params for Perceiver
    # byte_array and query will be automatically padded with zeros to get to d_model.
    # Set d_model to be divisible by `num_heads`.
    d_model: int = D_MODEL
    pv_system_id_embedding_dim: int = 16
    num_heads: int = N_HEADS
    dropout: float = 0.0
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 4

    def __post_init__(self):
        super().__init__()
        self.hrvsatellite_processor = HRVSatelliteProcessor()

        NUM_PV_SYSTEMS = 1400
        NUM_GSPS = 360
        id_embedding = nn.Embedding(
            num_embeddings=NUM_PV_SYSTEMS + NUM_GSPS,
            embedding_dim=self.pv_system_id_embedding_dim,
        )

        self.pv_query_generator = PVQueryGenerator(
            pv_system_id_embedding=id_embedding,
            t0_idx_5_min_training=T0_IDX_5_MIN_TRAINING,
            t0_idx_5_min_validation=T0_IDX_5_MIN_VALIDATION,
        )

        self.gsp_query_generator = GSPQueryGenerator(gsp_id_embedding=id_embedding)

        self.transformer_encoder = MultiLayerTransformerEncoder(
            d_model=self.d_model,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> dict[str, torch.Tensor]:
        # Reshape so each timestep is considered a different example:
        num_examples, num_timesteps = x[BatchKey.pv].shape[:2]
        # TODO: Fix this `original_x` hack, which is needed to prevent the reshaping affecting
        # the time_transformer, too!
        original_x = {}
        for batch_key in (
            BatchKey.gsp_5_min_time_utc_fourier,
            BatchKey.pv_time_utc_fourier,
            BatchKey.hrvsatellite_solar_azimuth,
            BatchKey.hrvsatellite_solar_elevation,
            BatchKey.hrvsatellite_actual,
            BatchKey.hrvsatellite_time_utc_fourier,
        ):
            original_x[batch_key] = x[batch_key]
            x[batch_key] = einops.rearrange(x[batch_key], "example time ... -> (example time) ...")

        # Process satellite data and queries:
        pv_query = self.pv_query_generator(x, for_satellite_transformer=True)
        gsp_query = self.gsp_query_generator(x, for_satellite_transformer=True)
        satellite_data = self.hrvsatellite_processor(x)

        # Pad with zeros if necessary to get up to self.d_model:
        pv_query = maybe_pad_with_zeros(pv_query, requested_dim=self.d_model)
        gsp_query = maybe_pad_with_zeros(gsp_query, requested_dim=self.d_model)
        satellite_data = maybe_pad_with_zeros(satellite_data, requested_dim=self.d_model)

        # Mask the NaN GSP queries. True or non-zero value indicates value will be ignored.
        mask = torch.concat(
            (
                torch.zeros_like(pv_query[:, :, 0]),
                torch.isnan(gsp_query[:, :, -1]),  # The time fourier features are last in dim 2.
                torch.zeros_like(satellite_data[:, :, 0]),
            ),
            dim=1,
        )

        # Now set NaN GSP queries to zero. They'll be ignored :)
        gsp_query = torch.nan_to_num(gsp_query, nan=0)

        # Prepare the attention input and run through the transformer_encoder:
        attn_input = torch.concat((pv_query, gsp_query, satellite_data), dim=1)
        attn_output = attn_input + self.transformer_encoder(attn_input, src_key_padding_mask=mask)

        # Reshape to (example time element d_model):
        attn_output = einops.rearrange(
            attn_output,
            "(example time) ... -> example time ...",
            example=num_examples,
            time=num_timesteps,
        )

        # Select the elements of the output which correspond to the query:
        gsp_start_idx = pv_query.shape[1]
        gsp_end_idx = gsp_start_idx + gsp_query.shape[1]
        pv_attn_out = attn_output[:, :, :gsp_start_idx]
        gsp_attn_out = attn_output[:, :, gsp_start_idx:gsp_end_idx]

        # Put back the original data! TODO: Remove this hack!
        x.update(original_x)

        return {
            "pv_attn_out": pv_attn_out,  # shape: (example, n_pv_systems, d_model)
            "gsp_attn_out": gsp_attn_out,  # shape: (example, 1, d_model)
        }


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class FullModel(pl.LightningModule):
    d_model: int = D_MODEL + N_HEADS  # + N_HEADS for historical PV
    pv_system_id_embedding_dim: int = 16
    num_heads: int = N_HEADS
    dropout: float = 0.1
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 4
    cheat: bool = False  #: Use real satellite imagery of the future.
    stop_gradients_before_unet: bool = True

    def __post_init__(self):
        super().__init__()

        if self.cheat:
            _log.warning("CHEATING MODE ENABLED! Using real satellite imagery of future!")

        # Predict future satellite images:
        self.satellite_predictor = SatellitePredictor.load_from_checkpoint(
            "/home/jack/model_params/022.25/exp_022.25_29_epochs.ckpt"
        )

        # Infer GSP and PV power output for a single timestep of satellite imagery.
        self.satellite_transformer = SatelliteTransformer()

        # Look across timesteps to produce the final output.
        self.time_transformer = MultiLayerTransformerEncoder(
            d_model=self.d_model,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

        self.pv_output_module = nn.Sequential(
            nn.Linear(in_features=self.d_model, out_features=self.d_model),
            nn.GELU(),
            nn.Linear(in_features=self.d_model, out_features=1),
            nn.ReLU(),  # ReLU output guarantees that we can't predict negative PV power!
        )

        # Do this at the end of __post_init__ to capture model topology to wandb:
        self.save_hyperparameters()

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> dict[str, torch.Tensor]:
        # Predict future satellite images.
        if not self.cheat:
            predicted_sat = self.satellite_predictor(x=x)  # Shape: example, time, y, x

            # Select a subset of predicted images, if we're in training mode:
            if BatchKey.requested_timesteps in x:
                # If `requested_timesteps` is in `x` then `ReduceNumTimesteps` was in the data
                # pipeline
                forecast_timesteps = x[BatchKey.requested_timesteps][NUM_HIST_SAT_IMAGES:]
                forecast_timesteps = forecast_timesteps - NUM_HIST_SAT_IMAGES
                predicted_sat = predicted_sat[:, forecast_timesteps]

            if self.stop_gradients_before_unet:
                predicted_sat = predicted_sat.detach()

            # Replace the "actual" future satellite images with predicted images
            # shape: (batch_size, time, channels, y, x, n_pixels_per_patch)
            x[BatchKey.hrvsatellite_actual][:, NUM_HIST_SAT_IMAGES:, 0] = predicted_sat

        sat_trans_out = self.satellite_transformer(x)
        pv_attn_out = sat_trans_out["pv_attn_out"]  # Shape: (example time n_pv_systems d_model)
        gsp_attn_out = sat_trans_out["gsp_attn_out"]

        # Concatenate actual historical PV on each PV attention output,
        # so the time_transformer doesn't have to put much effort into aligning
        # historical actual PV with predicted historical PV.
        t0_idx_5_min = T0_IDX_5_MIN_TRAINING if self.training else T0_IDX_5_MIN_VALIDATION
        historical_pv = torch.zeros_like(x[BatchKey.pv])  # Shape: (example, time, n_pv_systems)
        historical_pv[:, : t0_idx_5_min + 1] = x[BatchKey.pv][:, : t0_idx_5_min + 1]
        historical_pv = historical_pv.unsqueeze(-1)  # Shape: (example, time, n_pv_systems, 1)
        # Now append a "marker" to indicate which timesteps are history:
        hist_pv_marker = torch.zeros_like(historical_pv)
        hist_pv_marker[:, : t0_idx_5_min + 1] = 1
        pv_attn_out = torch.concat((pv_attn_out, historical_pv, hist_pv_marker), dim=3)

        # Reshape pv and gsp attention outputs so each timestep an each pv system is
        # seen as a separate element into the `time transformer`.
        n_timesteps, n_pv_systems = pv_attn_out.shape[1:3]
        REARRANGE_STR = "example time n_pv_systems d_model -> example (time n_pv_systems) d_model"
        pv_attn_out = einops.rearrange(pv_attn_out, REARRANGE_STR)
        pv_attn_out = maybe_pad_with_zeros(pv_attn_out, requested_dim=self.d_model)
        gsp_attn_out = einops.rearrange(gsp_attn_out, REARRANGE_STR, n_pv_systems=1)
        gsp_attn_out = maybe_pad_with_zeros(gsp_attn_out, requested_dim=self.d_model)
        n_pv_elements = pv_attn_out.shape[1]

        # Get GSP query
        gsp_query_generator = self.satellite_transformer.gsp_query_generator
        gsp_query = gsp_query_generator(x, for_satellite_transformer=False)
        gsp_query = maybe_pad_with_zeros(gsp_query, requested_dim=self.d_model)

        # Concatenate all the things we're going to feed into the "time transformer":
        time_attn_in = (pv_attn_out, gsp_attn_out, gsp_query)
        time_attn_in = torch.concat(time_attn_in, dim=1)
        time_attn_out = self.time_transformer(time_attn_in)

        power_out = self.pv_output_module(time_attn_out)  # (example, total_num_elements, 1)

        # Reshape the PV power predictions
        predicted_pv_power = power_out[:, :n_pv_elements]
        predicted_pv_power = einops.rearrange(
            predicted_pv_power,
            "example (time n_pv_systems) 1 -> example time n_pv_systems",
            time=n_timesteps,
            n_pv_systems=n_pv_systems,
        )

        # GSP power. There's just 1 GSP.
        n_gsp_elements = gsp_query.shape[1]
        predicted_gsp_power = power_out[:, -n_gsp_elements:, 0]

        return dict(
            predicted_pv_power=predicted_pv_power,  # Shape: (example time n_pv_systems)
            predicted_gsp_power=predicted_gsp_power,  # Shape: (example, time)
        )

    def training_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self.validation_step(batch, batch_idx)

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        tag = "train" if self.training else "validation"
        network_out = self(batch)
        predicted_pv_power = network_out["predicted_pv_power"]
        predicted_gsp_power = network_out["predicted_gsp_power"]
        actual_pv_power = batch[BatchKey.pv]
        actual_pv_power = torch.where(
            batch[BatchKey.pv_mask].unsqueeze(1),
            actual_pv_power,
            torch.tensor(0.0, dtype=actual_pv_power.dtype, device=actual_pv_power.device),
        )

        # PV power loss:
        pv_mse_loss = F.mse_loss(predicted_pv_power, actual_pv_power)
        t0_idx_5_min = T0_IDX_5_MIN_TRAINING if self.training else T0_IDX_5_MIN_VALIDATION

        pv_nmae_loss = F.l1_loss(
            predicted_pv_power[:, t0_idx_5_min + 1 :], actual_pv_power[:, t0_idx_5_min + 1 :]
        )
        self.log(f"{tag}/pv_mse", pv_mse_loss)
        self.log(f"{tag}/pv_nmae", pv_nmae_loss)

        # GSP power loss:
        actual_gsp_power = batch[BatchKey.gsp]
        gsp_mse_loss = F.mse_loss(predicted_gsp_power, actual_gsp_power)
        gsp_nmae_loss = F.l1_loss(
            predicted_gsp_power[:, T0_IDX_30_MIN + 1 :], actual_gsp_power[:, T0_IDX_30_MIN + 1 :]
        )
        self.log(f"{tag}/gsp_mse", gsp_mse_loss)
        self.log(f"{tag}/gsp_nmae", gsp_nmae_loss)

        # Total MSE loss:
        total_mse_loss = pv_mse_loss + gsp_mse_loss
        self.log(f"{tag}/total_mse", total_mse_loss)

        # Total NMAE loss:
        total_nmae_loss = pv_nmae_loss + gsp_nmae_loss
        self.log(f"{tag}/total_nmae", total_nmae_loss)

        return {
            "loss": total_mse_loss,
            "pv_mse_loss": pv_mse_loss,
            "gsp_mse_loss": gsp_mse_loss,
            "pv_nmae_loss": pv_nmae_loss,
            "gsp_nmae_loss": gsp_nmae_loss,
            "predicted_gsp_power": predicted_gsp_power,
            "actual_gsp_power": actual_gsp_power,
            "gsp_time_utc": batch[BatchKey.gsp_time_utc],
            "actual_pv_power": actual_pv_power,
            "predicted_pv_power": predicted_pv_power,
        }

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=5e-5)
        return optimizer


# ---------------------------------- Training ----------------------------------

model = FullModel()

wandb_logger = WandbLogger(
    name="023.02: Stop gradient before U-Net. U-Net & Power Perceiver. GCP-3",
    project="power_perceiver",
    entity="openclimatefix",
    log_model="all",
)

# log model only if validation loss decreases
checkpoint_callback = pl.callbacks.ModelCheckpoint(monitor="validation/total_mse", mode="min")


if socket.gethostname() == "donatello":
    GPU = 0
else:  # On GCP
    GPU = 0
trainer = pl.Trainer(
    gpus=[GPU],
    max_epochs=200,
    logger=wandb_logger,
    callbacks=[
        # LogSatellitePlots(),
        checkpoint_callback,
        pl.callbacks.LearningRateMonitor(logging_interval="step"),
        LogTimeseriesPlots(),
        LogTSNEPlot(query_generator_name="satellite_transformer.pv_query_generator"),
    ],
)

trainer.fit(
    model=model,
    train_dataloaders=train_dataloader,
    val_dataloaders=val_dataloader,
)



from power_perceiver.np_batch_processor.encode_space_time import EncodeSpaceTime
from power_perceiver.np_batch_processor.sun_position import SunPosition
from power_perceiver.np_batch_processor.topography import Topography



from dataclasses import dataclass
from typing import Iterable, Optional

import numpy as np

from power_perceiver.consts import BatchKey
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import (
    PreparedDataSource,
    XarrayBatch,
)


@dataclass
class ReduceNumTimesteps:
    """Reduce the number of timesteps per example to `requested_timesteps`.

    NOT ACTUALLY USED BY THE NEW TRAINING REGIME. INSTEAD WE SUBSELECT TIMESTEPS
    WITHIN THE MODEL ITSELF. THIS FOR TWO REASONS: 1) IT'S TRICKY TO USE
    XR_BATCH_PROCESSORS IN THE RawDataset. 2) WE WANT THE SatellitePredictor
    TO ALWAYS USE THE FULL SET OF TIMESTEPS.

    If `requested_timesteps` is an int then randomly pick `requested_timesteps` different
    timesteps for each batch.

    If `requested_timesteps` is an array of ints then use that array as the index into
    each xr_dataset.
    """

    num_requested_history_timesteps: int = 7  # Has to be total number available when using UNet.
    num_requested_forecast_timesteps: int = 8

    num_history_timesteps_available: int = 7
    num_total_timesteps_available: int = 31

    keys: Optional[Iterable[PreparedDataSource]] = None

    def __post_init__(self):
        # Any xr_batch_processor with an `rng` attribute will have the
        # rng seeded by the `seed_rngs` `worker_init_function`.
        self.rng = np.random.default_rng()

    def __call__(self, xr_batch: XarrayBatch) -> XarrayBatch:
        num_total_requested_timesteps = (
            self.num_requested_history_timesteps + self.num_requested_forecast_timesteps
        )
        # Need to use int64 so PyTorch can use requested_timesteps as an index.
        requested_timesteps = np.empty(num_total_requested_timesteps, dtype=np.int64)
        requested_timesteps[
            : self.num_requested_history_timesteps
        ] = random_int_without_replacement(
            start=0,
            stop=self.num_history_timesteps_available,
            num=self.num_requested_history_timesteps,
            rng=self.rng,
        )

        requested_timesteps[
            self.num_requested_history_timesteps :
        ] = random_int_without_replacement(
            start=self.num_history_timesteps_available,
            stop=self.num_total_timesteps_available,
            num=self.num_requested_forecast_timesteps,
            rng=self.rng,
        )

        if self.keys is None:
            keys = xr_batch.keys()
        else:
            keys = self.keys

        for key in keys:
            xr_dataset = xr_batch[key]
            xr_batch[key] = xr_dataset.isel(time=requested_timesteps)

        xr_batch[BatchKey.requested_timesteps] = requested_timesteps

        return xr_batch


def random_int_without_replacement(
    start: int, stop: int, num: int, rng: Optional[np.random.Generator] = None
) -> np.ndarray:
    """This seems to be the best way to get random ints *without* replacement"""
    if rng is None:
        rng = np.random.default_rng()
    ints = rng.choice(np.arange(start=start, stop=stop), size=num, replace=False)
    return np.sort(ints)



import datetime
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, ClassVar, Iterable, Optional

import fsspec
import numpy as np
import xarray as xr

from power_perceiver.consts import BatchKey

_log = logging.getLogger(__name__)


NumpyBatch = dict[BatchKey, np.ndarray]


@dataclass
class PreparedDataSource:
    """Load each data source.

    Initialisation arguments:
        data_path (Path): Optional.
        filename_suffix (str): Without the period (.)
        transforms: A list of transform functions. Each must accept an xr.Dataset, and must
            return a transformed xr.Dataset.

    How to add a new subclass:
      1. Create new subclass :)
      2. Override / set:
         - PreparedDataSource.to_numpy
      3. If necessary, also update BatchKey in consts.py
    """

    history_duration: datetime.timedelta
    sample_period_duration: ClassVar[datetime.timedelta]
    data_path: Optional[Path] = None
    filename_suffix: str = "nc"
    transforms: Optional[Iterable[Callable]] = None

    @classmethod
    @property
    def name(cls) -> str:
        return cls.__name__.lower()

    @property
    def full_data_path(self) -> Path:
        try:
            return self.data_path / self.name
        except:  # noqa: E722
            if self.data_path is None:
                raise ValueError("data_path must be set!")
            else:
                raise

    def __getitem__(self, batch_idx: int) -> xr.Dataset:
        filename = self.get_filename(batch_idx=batch_idx)
        dataset = load_netcdf(filename)
        dataset = self.process_before_transforms(dataset)
        if self.transforms:
            for transform in self.transforms:
                dataset = transform(dataset)
        return dataset

    def get_filename(self, batch_idx: int) -> Path:
        return self.full_data_path / f"{batch_idx:06d}.{self.filename_suffix}"

    def get_n_batches_available(self) -> int:
        n_batches = len(list(self.full_data_path.glob(f"*.{self.filename_suffix}")))
        _log.info(f"{self.name} has {n_batches} batches.")
        return n_batches

    def process_before_transforms(self, dataset: xr.Dataset) -> xr.Dataset:
        """Can be overridden by subclass."""
        return dataset

    @property
    def t0_idx(self) -> int:
        """The index into the array for the most recent observation (t0).

        Remember that, in this code, we consider t0 to be part of the history.
        So, if the history_duration is 1 hour, and sample_period_duration is 30 minutes,
        then "history" will be at indicies 0, 1, and 2.
        """
        return int(self.history_duration / self.sample_period_duration)

    @staticmethod
    def to_numpy(dataset: xr.Dataset) -> NumpyBatch:
        """This is called from Dataset._xarray_to_numpy_batch.

        Process this modality's xr.Dataset, to convert the xr.Dataset
        into a dictionary mapping BatchKeys to numpy arrays, as documented
        in the BatchKey class.
        """
        raise NotImplementedError


def load_netcdf(filename, engine="h5netcdf", *args, **kwargs) -> xr.Dataset:
    """Load a NetCDF dataset from local file system or cloud bucket."""
    with fsspec.open(filename, mode="rb") as file:
        dataset = xr.load_dataset(file, engine=engine, *args, **kwargs)
    return dataset


# The one exception to this type, is that `ReduceNumTimesteps`
# adds a `BatchKey.requested_timesteps` key, whose value is a np.ndarray.
XarrayBatch = dict[PreparedDataSource.__class__, xr.Dataset]



import datetime
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Union

import dask
import numpy as np
import pandas as pd
import pyproj
import pyresample
import xarray as xr

from power_perceiver.consts import BatchKey, Location
from power_perceiver.geospatial import OSGB36
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import NumpyBatch
from power_perceiver.load_prepared_batches.data_sources.satellite import SAT_MEAN, SAT_STD
from power_perceiver.load_raw.data_sources.raw_data_source import (
    RawDataSource,
    SpatialDataSource,
    TimeseriesDataSource,
    ZarrDataSource,
)
from power_perceiver.time import date_summary_str, select_data_in_daylight
from power_perceiver.utils import datetime64_to_float, is_sorted, select_time_periods

_log = logging.getLogger(__name__)


@dataclass(kw_only=True)
class RawSatelliteDataSource(
    # Surprisingly, Python's class hierarchy is defined right-to-left.
    # So the base class must go on the right.
    ZarrDataSource,
    TimeseriesDataSource,
    SpatialDataSource,
    RawDataSource,
):
    """Load satellite data directly from the intermediate satellite Zarr store."""

    _y_dim_name: ClassVar[str] = "y_geostationary"
    _x_dim_name: ClassVar[str] = "x_geostationary"
    # For now, let's assume the satellite imagery is always 5-minutely.
    # Later (WP3?), we'll want to experiment with lower temporal resolution satellite imagery.
    sample_period_duration: ClassVar[datetime.timedelta] = datetime.timedelta(minutes=5)
    needs_to_load_subset_into_ram: ClassVar[bool] = True

    def __post_init__(self):  # noqa: D105
        RawDataSource.__post_init__(self)
        SpatialDataSource.__post_init__(self)
        TimeseriesDataSource.__post_init__(self)
        ZarrDataSource.__post_init__(self)

    def per_worker_init(self, *args, **kwargs) -> None:  # noqa: D102
        super().per_worker_init(*args, **kwargs)
        self.open()

    def open(self) -> None:
        """
        Lazily open Satellite data (this only loads metadata. It doesn't load all the data!)

        We don't want to call `open_sat_data` in __init__.
        If we did that, then we couldn't copy RawSatelliteDataSource
        instances into separate processes.  Instead,
        call open() _after_ creating separate processes.
        """
        self._data_on_disk = open_sat_data(zarr_path=self.zarr_path)

        self._load_geostationary_area_definition_and_transform()

        # Check the x and y coords are sorted. If they are not then searchsorted won't work!
        assert is_sorted(self.data_on_disk[self._y_dim_name][::-1])
        assert is_sorted(self.data_on_disk[self._x_dim_name])

        # Sub-select data:
        _log.info("Before any selection: " + date_summary_str(self.data_on_disk.time_utc))

        # Select only the timesteps we want:
        self._data_on_disk = select_time_periods(
            xr_data=self.data_on_disk,
            time_periods=self.time_periods,
        )
        self._data_on_disk = select_data_in_daylight(self._data_on_disk)
        _log.info("After filtering: " + date_summary_str(self.data_on_disk.time_utc))

    def _post_process(self, xr_data: xr.DataArray) -> xr.DataArray:
        # hrvsatellite is int16 on disk but, upstream of this method, we convert to uint8 to save
        # RAM. We convert to uint8 by dividing by 4 (to get from pixel values in the range
        # [0, 1023] to the range [0, 255]). So `xr_data` will be uint8 when it gets to this method.
        # But `SAT_MEAN` and `SAT_STD` were computed on the original satellite data,
        # which is in the range [0, 1023] (i.e. 4 times greater than the uint8 range of [0, 255])
        # We divide the mean and std by 4 because the `xr_data` has also been divided by 4.
        xr_data = xr_data.astype(np.float32)
        xr_data = xr_data - (SAT_MEAN["HRV"] / 4)
        xr_data = xr_data / (SAT_STD["HRV"] / 4)
        return xr_data

    @staticmethod
    def to_numpy(xr_data: xr.DataArray) -> NumpyBatch:
        """Convert xarray to numpy batch.

        But note that this is actually just returns *one* example (not a whole batch!)
        """
        example: NumpyBatch = {}

        example[BatchKey.hrvsatellite_actual] = xr_data.values
        example[BatchKey.hrvsatellite_t0_idx] = xr_data.attrs["t0_idx"]
        example[BatchKey.hrvsatellite_time_utc] = datetime64_to_float(xr_data["time_utc"].values)
        for batch_key, dataset_key in (
            (BatchKey.hrvsatellite_y_osgb, "y_osgb"),
            (BatchKey.hrvsatellite_x_osgb, "x_osgb"),
            (BatchKey.hrvsatellite_y_geostationary, RawSatelliteDataSource._y_dim_name),
            (BatchKey.hrvsatellite_x_geostationary, RawSatelliteDataSource._x_dim_name),
        ):
            # HRVSatellite coords are already float32.
            example[batch_key] = xr_data[dataset_key].values

        return example

    def _load_geostationary_area_definition_and_transform(self) -> None:
        area_definition_yaml = self.data_on_disk.attrs["area"]
        geostationary_area_definition = pyresample.area_config.load_area_from_string(
            area_definition_yaml
        )
        geostationary_crs = geostationary_area_definition.crs
        self._osgb_to_geostationary = pyproj.Transformer.from_crs(
            crs_from=OSGB36, crs_to=geostationary_crs
        ).transform

    def _get_idx_of_pixel_at_center_of_roi(
        self, xr_data: xr.DataArray, center_osgb: Location
    ) -> Location:
        """Return x and y index location of pixel at center of region of interest."""
        center_geostationary_tuple = self._osgb_to_geostationary(xx=center_osgb.x, yy=center_osgb.y)
        center_geostationary = Location(
            x=center_geostationary_tuple[0], y=center_geostationary_tuple[1]
        )

        # Get the index into x and y nearest to x_center_geostationary and y_center_geostationary:
        x_index_at_center = (
            np.searchsorted(xr_data[self._x_dim_name].values, center_geostationary.x) - 1
        )
        # y_geostationary is in descending order:
        y_index_at_center = len(xr_data[self._y_dim_name]) - (
            np.searchsorted(xr_data[self._y_dim_name].values[::-1], center_geostationary.y) - 1
        )
        return Location(x=x_index_at_center, y=y_index_at_center)


def open_sat_data(
    zarr_path: Union[Path, str],
    convert_to_uint8: bool = True,
) -> xr.DataArray:
    """Lazily opens the Zarr store.

    Args:
      zarr_path: Cloud URL or local path pattern.  If GCP URL, must start with 'gs://'
    """
    _log.debug("Opening satellite data: %s", zarr_path)

    # Silence the warning about large chunks.
    # Alternatively, we could set this to True, but that slows down loading a Satellite batch
    # from 8 seconds to 50 seconds!
    dask.config.set({"array.slicing.split_large_chunks": False})

    # Open the data
    dataset = xr.open_dataset(zarr_path, engine="zarr", chunks="auto")

    # Flip coordinates to top-left first
    dataset = dataset.reindex(y=dataset.y[::-1])
    dataset = dataset.reindex(x=dataset.x[::-1])

    # Rename
    # These renamings will no longer be necessary when the Zarr uses the 'correct' names,
    # see https://github.com/openclimatefix/Satip/issues/66
    if "variable" in dataset:
        dataset = dataset.rename({"variable": "channel"})
    elif "channel" not in dataset:
        # This is HRV version 3, which doesn't have a channels dim.  So add one.
        dataset = dataset.expand_dims(dim={"channel": ["HRV"]}, axis=1)

    # Rename coords to be more explicit about exactly what some coordinates hold:
    # Note that `rename` renames *both* the coordinates and dimensions, and keeps
    # the connection between the dims and coordinates, so we don't have to manually
    # use `data_array.set_index()`.
    dataset = dataset.rename(
        {
            "time": "time_utc",
            "y": "y_geostationary",
            "x": "x_geostationary",
        }
    )

    data_array = dataset["data"]
    del dataset

    # Ensure the y and x coords are in the right order (top-left first):
    assert data_array.y_geostationary[0] > data_array.y_geostationary[-1]
    assert data_array.x_geostationary[0] < data_array.x_geostationary[-1]
    assert data_array.y_osgb[0, 0] > data_array.y_osgb[-1, 0]
    assert data_array.x_osgb[0, 0] < data_array.x_osgb[0, -1]

    if convert_to_uint8:
        data_array = data_array.clip(min=0, max=1023)
        data_array.data = (data_array.astype(np.float32).data / 4.0).round().astype(np.uint8)

    # Sanity checks!
    assert data_array.dims == ("time_utc", "channel", "y_geostationary", "x_geostationary")
    datetime_index = pd.DatetimeIndex(data_array.time_utc)
    assert datetime_index.is_unique
    assert datetime_index.is_monotonic_increasing
    # Satellite datetimes can sometimes be 04, 09, minutes past the hour, or other slight offsets.
    # These slight offsets will break downstream code, which expects satellite data to be at
    # exactly 5 minutes past the hour.
    assert (datetime_index == datetime_index.round("5T")).all()

    return data_array



import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Iterable, Optional

import numpy as np
import torch
import torch.utils.data

from power_perceiver.consts import BatchKey
from power_perceiver.load_prepared_batches.data_sources import NumpyBatch, PreparedDataSource
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import XarrayBatch
from power_perceiver.np_batch_processor import EncodeSpaceTime, SunPosition, Topography
from power_perceiver.np_batch_processor.align_gsp_to_5_min import AlignGSPTo5Min
from power_perceiver.np_batch_processor.save_t0_time import SaveT0Time
from power_perceiver.testing import get_path_of_local_topo_data

_log = logging.getLogger(__name__)


@dataclass
class PreparedDataset(torch.utils.data.Dataset):
    """Load batches pre-prepared by `nowcasting_dataset`.

    Initialisation arguments:
        data_loaders: A list of instantiated data loader objects.
        data_path: Base path to the pre-prepared dataset. e.g. /path/to/v15/train/
        max_n_batches_per_epoch: If the user sets this to an int then
            this int will be the max number of batches used per epoch. If left as None
            then will load as many batches as are available.
        xr_batch_processors: Functions which takes an XarrayBatch,
            and does processing *across* modalities, and returns the processed XarrayBatch.
            Note that and processing *within* a modality should be done in
            PreparedDataSource.to_numpy.
        np_batch_processors: Functions which takes a NumpyBatch,
            and does processing *across* modalities, and returns the processed NumpyBatch.
            Note that and processing *within* a modality should be done in
            PreparedDataSource.to_numpy.

    Attributes:
        n_batches: int. Set by _set_number_of_batches.
    """

    data_loaders: Iterable[PreparedDataSource]
    data_path: Optional[Path] = None
    max_n_batches_per_epoch: Optional[int] = None
    xr_batch_processors: Optional[Iterable[Callable]] = None
    np_batch_processors: Optional[Iterable[Callable]] = None
    topography_location: str = str(
        "/home/jacob/Development/power_perceiver/data/europe_dem_2km_osgb.tif"
    )

    def __post_init__(self):
        # Sanity checks
        assert self.data_path.exists()
        assert len(self.data_loaders) > 0
        # Prepare PreparedDataSources.
        self._set_data_path_in_data_loaders()
        self._set_number_of_batches()
        np_batch_processors = [AlignGSPTo5Min(), EncodeSpaceTime(), SaveT0Time()]
        for modality_name in ["hrvsatellite", "gsp", "gsp_5_min", "pv", "nwp_target_time"]:
            np_batch_processors.append(SunPosition(modality_name=modality_name))
        #np_batch_processors.append(Topography(self.topography_location))
        self.np_batch_processors = np_batch_processors
        super().__init__()

    def _set_data_path_in_data_loaders(self) -> None:
        for data_loader in self.data_loaders:
            data_loader.data_path = self.data_path

    def _set_number_of_batches(self) -> None:
        """Set number of batches.  Check every data source."""
        self.n_batches = None
        for data_loader in self.data_loaders:
            n_batches_for_data_source = data_loader.get_n_batches_available()
            if self.n_batches is None:
                self.n_batches = n_batches_for_data_source
            elif n_batches_for_data_source != self.n_batches:
                self.n_batches = min(self.n_batches, n_batches_for_data_source)
                _log.warning(
                    f"Warning! {data_loader} has a different number of batches to at"
                    " least one other modality!"
                    f" We'll use the minimum of the two values: {self.n_batches}"
                )
        if self.max_n_batches_per_epoch is not None:
            self.n_batches = min(self.n_batches, self.max_n_batches_per_epoch)
        assert self.n_batches is not None
        assert self.n_batches > 0

    def __len__(self) -> int:
        return self.n_batches

    def __getitem__(self, batch_idx: int) -> NumpyBatch:
        if batch_idx >= self.n_batches:
            raise KeyError(f"{batch_idx=} is out of bounds! {self.n_batches=}")
        xr_batch = self._get_xarray_batch(batch_idx)
        xr_batch = self._process_xr_batch(xr_batch)
        np_batch = self._xarray_to_numpy_batch(xr_batch)
        del xr_batch
        np_batch = self._process_np_batch(np_batch)
        np_batch[BatchKey.hrvsatellite_surface_height] = np.zeros_like(np_batch[BatchKey.hrvsatellite_actual][:,0,0,:])
        return np_batch

    def _get_xarray_batch(self, batch_idx: int) -> NumpyBatch:
        """Load the completely un-modified batches from disk and store them in a dict."""
        xr_batch: NumpyBatch = {}
        for data_loader in self.data_loaders:
            xr_data_for_data_source = data_loader[batch_idx]
            xr_batch[data_loader.__class__] = xr_data_for_data_source
        return xr_batch

    def _process_xr_batch(self, xr_batch: XarrayBatch) -> XarrayBatch:
        """If necessary, do any processing which needs to be done across modalities,
        on the xr.Datasets."""
        if self.xr_batch_processors:
            for xr_batch_processor in self.xr_batch_processors:
                xr_batch = xr_batch_processor(xr_batch)
        return xr_batch

    def _xarray_to_numpy_batch(self, xr_batch: XarrayBatch) -> NumpyBatch:
        """Convert from xarray Datasets to numpy."""
        np_batch: NumpyBatch = {}
        for data_loader_class, xr_dataset in xr_batch.items():
            if data_loader_class == BatchKey.requested_timesteps:
                # `ReduceNumTimesteps` introduces a `requested_timesteps` key,
                # whose value is a np.ndarray.
                requested_timesteps = xr_dataset
                np_batch[BatchKey.requested_timesteps] = requested_timesteps
            else:
                np_data_for_data_source = data_loader_class.to_numpy(xr_dataset)
                np_batch.update(np_data_for_data_source)
        return np_batch

    def _process_np_batch(self, np_batch: NumpyBatch) -> NumpyBatch:
        """If necessary, do any processing which needs to be done across modalities,
        on the NumpyBatch."""
        if self.np_batch_processors:
            for np_batch_processor in self.np_batch_processors:
                np_batch = np_batch_processor(np_batch)
        return np_batch



"""
U-Net followed by the Power Perceiver.

See this issue for a diagram: https://github.com/openclimatefix/power_perceiver/issues/54
"""

# General imports
import logging
from dataclasses import dataclass
from typing import Optional

import einops

# ML imports
import matplotlib.pyplot as plt
import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from torch import nn

# power_perceiver imports
from power_perceiver.consts import BatchKey
from power_perceiver.hub import NowcastingModelHubMixin
from power_perceiver.production.pvrnn import N_PV_SYSTEMS_PER_EXAMPLE, PVRNN
from power_perceiver.production.satellite_transformer import (
    BOTTOM_IDX,
    D_MODEL,
    LEFT_IDX,
    N_HEADS,
    NUM_FUTURE_SAT_IMAGES,
    NUM_HIST_SAT_IMAGES,
    RIGHT_IDX,
    SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS,
    TOP_IDX,
    SatellitePredictor,
    SatelliteTransformer,
    maybe_pad_with_zeros,
)
from power_perceiver.pytorch_modules.mixture_density_network import (
    MixtureDensityNetwork,
    get_distribution,
)
from power_perceiver.pytorch_modules.nwp_processor import NWPProcessor
from power_perceiver.pytorch_modules.query_generator import GSPQueryGenerator
from power_perceiver.pytorch_modules.self_attention import MultiLayerTransformerEncoder
from power_perceiver.utils import assert_num_dims
from power_perceiver.xr_batch_processor.reduce_num_timesteps import random_int_without_replacement

logging.basicConfig()
_log = logging.getLogger("power_perceiver")
_log.setLevel(logging.DEBUG)

plt.rcParams["figure.figsize"] = (18, 10)
plt.rcParams["figure.facecolor"] = "white"

# SatellitePredictor options

# PowerPerceiver options

NWP_CHANNELS = ("t", "dswrf", "prate", "r", "si10", "vis", "lcc", "mcc", "hcc")

DEBUG = False
ENABLE_WANDB = False


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6


# ---------------------------------- SatelliteTransformer ----------------------------------


# Indexes for cropping the centre of the satellite image:


def _crop_satellite_and_spatial_coords(
    x: dict[BatchKey, torch.Tensor]
) -> dict[BatchKey, torch.Tensor]:
    """Crop satellite data & spatial coordinates.

    This is necessary because we want to give the "satellite predictor"
    a large rectangle of imagery (e.g. 256 wide x 128 high) so it can see clouds
    coming, but we probably don't want to give that huge image to a full-self-attention model.
    """
    assert_num_dims(x[BatchKey.hrvsatellite_actual], num_expected_dims=5)
    assert_num_dims(x[BatchKey.hrvsatellite_predicted], num_expected_dims=4)

    def _check_shape(batch_key, y_idx=1, x_idx=2):
        error_msg = f"{batch_key.name}.shape = {x[batch_key].shape}"
        assert x[batch_key].shape[y_idx] == SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS, error_msg
        assert x[batch_key].shape[x_idx] == SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS, error_msg

    # Crop the predicted and actual imagery:
    for batch_key in (BatchKey.hrvsatellite_actual, BatchKey.hrvsatellite_predicted):
        x[batch_key] = x[batch_key][..., TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX]
        _check_shape(batch_key, y_idx=-2, x_idx=-1)

    # Crop the coords (which have to be done in a separate loop, because `y` and `x`
    # are at different positions):
    for batch_key in (
        BatchKey.hrvsatellite_y_osgb_fourier,
        BatchKey.hrvsatellite_x_osgb_fourier,
        BatchKey.hrvsatellite_surface_height,
    ):
        x[batch_key] = x[batch_key][:, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX]
        _check_shape(batch_key)
    return x


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6


@dataclass(eq=False)
class FullModel(pl.LightningModule, NowcastingModelHubMixin):
    d_model: int = D_MODEL
    num_heads: int = N_HEADS
    dropout: float = 0.1
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 8
    cheat: bool = False  #: Use real satellite imagery of the future.
    #: Compute the loss on a central crop of the imagery.
    num_5_min_history_timesteps_during_training: Optional[int] = 4
    num_5_min_forecast_timesteps_during_training: Optional[int] = 6
    num_gaussians: int = 4
    num_rnn_layers: int = 4

    def __post_init__(self):
        super().__init__()

        if self.cheat:
            _log.warning("CHEATING MODE ENABLED! Using real satellite imagery of future!")

        # Predict future satellite images:
        self.satellite_predictor = SatellitePredictor()

        # Infer GSP and PV power output for a single timestep of satellite imagery.
        self.satellite_transformer = SatelliteTransformer()

        self.nwp_processor = NWPProcessor(n_channels=len(NWP_CHANNELS))

        # Find temporal features, and help calibrate predictions using recent history:
        self.pv_rnn = PVRNN(
            hidden_size=self.d_model,
            num_layers=self.num_rnn_layers,
            dropout=self.dropout,
        )

        # Look across timesteps to produce the final output.
        self.time_transformer = MultiLayerTransformerEncoder(
            d_model=self.d_model,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

        self.pv_mixture_density_net = nn.Sequential(
            nn.Linear(in_features=self.d_model, out_features=self.d_model),
            nn.GELU(),
            nn.Linear(in_features=self.d_model, out_features=self.d_model),
            nn.GELU(),
            MixtureDensityNetwork(in_features=self.d_model, num_gaussians=self.num_gaussians),
        )

        # Do this at the end of __post_init__ to capture model topology to wandb:
        self.save_hyperparameters()

    def _predict_satellite(self, x: dict[BatchKey, torch.Tensor]) -> dict[BatchKey, torch.Tensor]:
        """Predict future satellite images. The SatellitePredictor always gets every timestep."""
        if self.cheat:
            predicted = x[BatchKey.hrvsatellite_actual][:, :NUM_HIST_SAT_IMAGES, 0]
        else:
            predicted = self.satellite_predictor(x=x)  # Shape: example, time, y, x

        assert predicted.isfinite().all()
        x[BatchKey.hrvsatellite_predicted] = predicted
        return x

    def _select_random_subset_of_timesteps(
        self, x: dict[BatchKey, torch.Tensor]
    ) -> dict[BatchKey, torch.Tensor]:
        random_history_timestep_indexes = random_int_without_replacement(
            start=0,
            stop=NUM_HIST_SAT_IMAGES,
            num=self.num_5_min_history_timesteps_during_training,
        )
        random_forecast_timestep_indexes = random_int_without_replacement(
            start=NUM_HIST_SAT_IMAGES,
            stop=NUM_HIST_SAT_IMAGES + NUM_FUTURE_SAT_IMAGES,
            num=self.num_5_min_forecast_timesteps_during_training,
        )
        random_timestep_indexes = np.concatenate(
            (random_history_timestep_indexes, random_forecast_timestep_indexes)
        )
        x[BatchKey.hrvsatellite_actual] = x[BatchKey.hrvsatellite_actual][
            :, random_history_timestep_indexes
        ]
        x[BatchKey.hrvsatellite_predicted] = x[BatchKey.hrvsatellite_predicted][
            :, random_forecast_timestep_indexes - NUM_HIST_SAT_IMAGES
        ]
        for batch_key in (
            # If we go back to training the u-net, then we'll have to save `hrvsatellite`
            # before it is subsampled. See issue #156.
            # Don't include BatchKey.hrvsatellite_time_utc because all it's used for
            # is plotting satellite predictions, and we need all timesteps for that!
            # Subselect pv_time_utc here, so `plot_probabability_timeseries.plot_pv_power`
            # works correctly.
            # We *do* subset hrvsatellite_time_utc_fourier because it's used in the
            # satellite_transformer.
            BatchKey.hrvsatellite_time_utc_fourier,
            BatchKey.hrvsatellite_solar_azimuth,
            BatchKey.hrvsatellite_solar_elevation,
            BatchKey.pv,
            BatchKey.pv_time_utc,
            BatchKey.pv_time_utc_fourier,
            BatchKey.pv_solar_azimuth,
            BatchKey.pv_solar_elevation,
            BatchKey.gsp_5_min,
            BatchKey.gsp_5_min_time_utc_fourier,
            BatchKey.gsp_5_min_solar_azimuth,
            BatchKey.gsp_5_min_solar_elevation,
        ):
            x[batch_key] = x[batch_key][:, random_timestep_indexes]

        for batch_key in (
            BatchKey.hrvsatellite_t0_idx,
            BatchKey.pv_t0_idx,
            BatchKey.gsp_5_min_t0_idx,
        ):
            x[batch_key] = self.num_5_min_history_timesteps_during_training - 1

        return x

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> dict[str, torch.Tensor]:
        # Predict future satellite images. The SatellitePredictor always gets every timestep.
        x = self._predict_satellite(x)

        # Select a subset of 5-minute timesteps during training:
        if self.num_5_min_history_timesteps_during_training and self.training:
            x = self._select_random_subset_of_timesteps(x)
            num_5_min_timesteps = (
                self.num_5_min_history_timesteps_during_training
                + self.num_5_min_forecast_timesteps_during_training
            )
        else:
            num_5_min_timesteps = NUM_HIST_SAT_IMAGES + NUM_FUTURE_SAT_IMAGES

        x = _crop_satellite_and_spatial_coords(x)

        # Pass through the transformer!
        sat_trans_out = self.satellite_transformer(x=x)

        # Pass through PV RNN:
        sat_trans_pv_attn_out = sat_trans_out["pv_attn_out"]
        # Shape of `sat_trans_pv_attn_out`: (example time n_pv_systems d_model)
        pv_rnn_out = self.pv_rnn(x=x, sat_trans_pv_attn_out=sat_trans_pv_attn_out)
        del sat_trans_pv_attn_out
        n_pv_elements = pv_rnn_out.shape[1]

        # ----------------- Prepare inputs for the `time_transformer` -----------------------------

        # Get GSP attention outputs. Each GSP attention output timestep is seen as a separate
        # element into the `time transformer`. Remember, at this stage, `sat_trans_gsp_attn_out`
        # is 5-minutely. Shape (after squeeze): example time d_model.
        sat_trans_gsp_attn_out = sat_trans_out["gsp_attn_out"].squeeze()

        # Pad with zeros
        pv_rnn_out = maybe_pad_with_zeros(pv_rnn_out, requested_dim=self.d_model)
        sat_trans_gsp_attn_out = maybe_pad_with_zeros(
            sat_trans_gsp_attn_out, requested_dim=self.d_model
        )

        # Get GSP query for the time_transformer:
        # The query for the time_transformer is just for the half-hourly timesteps
        # (not the 5-minutely GSP queries used in the `SatelliteTransformer`.)
        gsp_query_generator: GSPQueryGenerator = self.satellite_transformer.gsp_query_generator
        gsp_query = gsp_query_generator.forward(
            x,
            include_history=True,
            base_batch_key=BatchKey.gsp,
            do_reshape_time_as_batch=False,
        )
        gsp_query = maybe_pad_with_zeros(gsp_query, requested_dim=self.d_model)

        # Prepare NWP inputs
        nwp_query = self.nwp_processor(x)
        nwp_query = maybe_pad_with_zeros(nwp_query, requested_dim=self.d_model)

        # Concatenate all the things we're going to feed into the "time transformer":
        # Shape: (example, elements, d_model)
        # `pv_rnn_out` must be the first set of elements.
        # `gsp_query` must be the last set of elements.
        # The shape is (example, query_elements, d_model).
        time_attn_in = torch.concat(
            (pv_rnn_out, sat_trans_gsp_attn_out, nwp_query, gsp_query), dim=1
        )

        # It's necessary to mask the time_transformer because some examples, where we're trying to
        # predict GSP PV power, might have NaN PV data (because the GSP is so far north it has no
        # PV systems!), and we don't want the attention mechanism to pay any attention to the
        # zero'd out (previously NaN) PV data when predicting GSP PV power.
        mask = time_attn_in.isnan().any(dim=2)
        time_attn_in = time_attn_in.nan_to_num(0)
        time_attn_out = self.time_transformer(time_attn_in, src_key_padding_mask=mask)

        # ---------------------------- MIXTURE DENSITY NETWORK -----------------------------------
        # The MDN doesn't like NaNs, and I think there will be NaNs when, for example,
        # the example is missing PV data:
        time_attn_out = time_attn_out.nan_to_num(0)
        predicted_pv_power = self.pv_mixture_density_net(time_attn_out[:, :n_pv_elements])

        # Reshape the PV power predictions
        predicted_pv_power = einops.rearrange(
            predicted_pv_power,
            "example (time n_pv_systems) mdn_features -> example time n_pv_systems mdn_features",
            time=num_5_min_timesteps,
            n_pv_systems=N_PV_SYSTEMS_PER_EXAMPLE,
            mdn_features=self.num_gaussians * 3,  # x3 for pi, mu, sigma.
        )

        # GSP power. There's just 1 GSP. So each gsp element is a timestep.
        n_gsp_elements = gsp_query.shape[1]
        predicted_gsp_power = self.pv_mixture_density_net(time_attn_out[:, -n_gsp_elements:])

        return dict(
            predicted_pv_power=predicted_pv_power,  # Shape: (example time n_pv_sys mdn_features)
            predicted_gsp_power=predicted_gsp_power,  # Shape: (example time mdn_features)
            predicted_sat=x[BatchKey.hrvsatellite_predicted],  # Shape: example, time, y, x
            pv_time_utc=x[BatchKey.pv_time_utc],  # Give the subsampled times to plotting functions.
        )

    @property
    def tag(self) -> str:
        return "train" if self.training else "validation"

    def training_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self.validation_step(batch, batch_idx)

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:

        network_out = self(batch)

        # PV & GSP LOSS ######################
        predicted_pv_power = network_out["predicted_pv_power"]
        actual_pv_power = batch[BatchKey.pv]
        predicted_gsp_power = network_out["predicted_gsp_power"]
        actual_gsp_power = batch[BatchKey.gsp].squeeze()

        # Mask predicted and actual PV and GSP (some examples don't have PV and/or GSP)
        # For more discussion of how to mask losses in pytorch, see:
        # https://discuss.pytorch.org/t/masking-input-to-loss-function/121830/3
        pv_mask = actual_pv_power.isfinite()
        pv_t0_idx = batch[BatchKey.pv_t0_idx]
        pv_slice_from_t0 = slice(pv_t0_idx + 1, None)
        pv_mask_from_t0 = pv_mask[:, pv_slice_from_t0]

        gsp_mask = actual_gsp_power.isfinite()
        gsp_t0_idx = batch[BatchKey.gsp_t0_idx]
        gsp_slice_from_t0 = slice(gsp_t0_idx + 1, None)
        gsp_slice_from_t0_to_1h = slice(gsp_t0_idx + 1, gsp_t0_idx + 3)
        gsp_mask_from_t0 = gsp_mask[:, gsp_slice_from_t0]
        gsp_mask_from_t0_to_1h = gsp_mask[:, gsp_slice_from_t0_to_1h]

        predicted_pv_power_from_t0 = predicted_pv_power[:, pv_slice_from_t0][pv_mask_from_t0]
        actual_pv_power_from_t0 = actual_pv_power[:, pv_slice_from_t0][pv_mask_from_t0]

        predicted_gsp_power_from_t0 = predicted_gsp_power[:, gsp_slice_from_t0][gsp_mask_from_t0]
        actual_gsp_power_from_t0 = actual_gsp_power[:, gsp_slice_from_t0][gsp_mask_from_t0]

        predicted_gsp_power_from_t0_to_1h = predicted_gsp_power[:, gsp_slice_from_t0_to_1h][
            gsp_mask_from_t0_to_1h
        ]
        actual_gsp_power_from_t0_to_1h = actual_gsp_power[:, gsp_slice_from_t0_to_1h][
            gsp_mask_from_t0_to_1h
        ]

        # PV negative log prob loss:
        pv_distribution = get_distribution(predicted_pv_power_from_t0)
        pv_neg_log_prob_loss = -pv_distribution.log_prob(actual_pv_power_from_t0).mean()
        self.log(f"{self.tag}/pv_neg_log_prob", pv_neg_log_prob_loss)

        # PV power loss:
        pv_mse_loss = F.mse_loss(pv_distribution.mean, actual_pv_power_from_t0)
        self.log(f"{self.tag}/pv_mse", pv_mse_loss)

        pv_nmae_loss = F.l1_loss(pv_distribution.mean, actual_pv_power_from_t0)
        self.log(f"{self.tag}/pv_nmae", pv_nmae_loss)

        # GSP negative log prob loss from t0:
        gsp_distribution_from_t0 = get_distribution(predicted_gsp_power_from_t0)
        gsp_neg_log_prob_loss_from_t0 = -gsp_distribution_from_t0.log_prob(
            actual_gsp_power_from_t0
        ).mean()
        self.log(f"{self.tag}/gsp_neg_log_prob", gsp_neg_log_prob_loss_from_t0)

        # GSP negative log prob loss from t0 to 1h forecast:
        gsp_distribution_from_t0_to_1h = get_distribution(predicted_gsp_power_from_t0_to_1h)
        gsp_neg_log_prob_loss_from_t0_to_1h = -gsp_distribution_from_t0_to_1h.log_prob(
            actual_gsp_power_from_t0_to_1h
        ).mean()
        self.log(f"{self.tag}/gsp_neg_log_prob_from_t0_to_1h", gsp_neg_log_prob_loss_from_t0_to_1h)

        # GSP power loss:
        gsp_mse_loss = F.mse_loss(gsp_distribution_from_t0.mean, actual_gsp_power_from_t0)
        self.log(f"{self.tag}/gsp_mse", gsp_mse_loss)

        gsp_nmae_loss = F.l1_loss(gsp_distribution_from_t0.mean, actual_gsp_power_from_t0)
        self.log(f"{self.tag}/gsp_nmae", gsp_nmae_loss)

        # Total PV and GSP MSE loss:
        total_pv_and_gsp_mse_loss = gsp_mse_loss + pv_mse_loss
        self.log(f"{self.tag}/total_mse", total_pv_and_gsp_mse_loss)

        # Total NMAE loss:
        total_nmae_loss = pv_nmae_loss + gsp_nmae_loss
        self.log(f"{self.tag}/total_nmae", total_nmae_loss)

        # Total PV and GSP negative log probability loss:
        total_pv_and_gsp_neg_log_prob_loss = (
            gsp_neg_log_prob_loss_from_t0
            + (gsp_neg_log_prob_loss_from_t0_to_1h * 2)
            + pv_neg_log_prob_loss
        )
        self.log(
            f"{self.tag}/total_pv_and_gsp_neg_log_prob_loss", total_pv_and_gsp_neg_log_prob_loss
        )

        return {
            "loss": total_pv_and_gsp_neg_log_prob_loss,
            "predicted_gsp_power": predicted_gsp_power,
            "predicted_gsp_power_mean": get_distribution(predicted_gsp_power).mean,
            "actual_gsp_power": actual_gsp_power,
            "gsp_time_utc": batch[BatchKey.gsp_time_utc],
            "actual_pv_power": actual_pv_power,
            "predicted_pv_power": predicted_pv_power,
            "predicted_pv_power_mean": get_distribution(predicted_pv_power).mean,
            "pv_time_utc": network_out["pv_time_utc"],
            "predicted_sat": network_out["predicted_sat"],  # Shape: example, time, y, x
            "actual_sat": batch[BatchKey.hrvsatellite_actual][:, NUM_HIST_SAT_IMAGES:, 0],
        }

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=5e-5)

        def _lr_lambda(epoch):
            return 50 / (epoch + 50)

        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, _lr_lambda, verbose=True)
        return [optimizer], [scheduler]
